#    Copyright (C) 2011 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Routines to export the document."""

import os.path
import random
import codecs
import re
import sys
import subprocess

from .. import qtall as qt
from .. import setting
from .. import utils

try:
    from . import emf_export
    hasemf = True
except ImportError:
    hasemf = False

from . import svg_export
from . import selftest_export
from . import painthelper

# 1m in inch
m_inch = 39.370079

# scale factor for svg dpi
svg_dpi_scale = 0.1

def _(text, disambiguation=None, context="Export"):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

def getSinglePage(pagenumbers):
    """Check single number of pages or throw exception, else return page number."""

    if len(pagenumbers) != 1:
        raise RuntimeError(
            'Can only export a single page in this format')
    return pagenumbers[0]

class ExportRunnable(qt.QRunnable):
    """For running export in thread."""

    def __init__(self, aexport, filename, phelpers):
        qt.QRunnable.__init__(self)
        self.aexport = aexport
        self.filename = filename
        self.phelpers = phelpers

    def run(self):
        """Do export.

        Exceptions are returned to AsyncExport class."""
        try:
            self.doExport()
        except (RuntimeError, EnvironmentError) as e:
            self.aexport.exception = e

    def renderPage(self, dev, phelper):
        """Render page, clipping."""
        painter = qt.QPainter(dev)
        painter.setClipRect(qt.QRectF(
            qt.QPointF(0,0), qt.QPointF(*phelper.pagesize)))
        painter.save()
        phelper.renderToPainter(painter)
        painter.restore()
        painter.end()

class ExportBitmapRunnable(ExportRunnable):
    """Runnable task to export a bitmap."""

    def doExport(self):
        """Do the export."""
        ext = os.path.splitext(self.filename)[1].lower()
        fmt = ext.lstrip('.') # setFormat() doesn't want the leading '.'
        if fmt == 'jpeg':
            fmt = 'jpg'

        # create real output image
        size = self.phelpers[0].pagesize
        backqcolor = self.aexport.backqcolor
        if fmt == 'png':
            # transparent output
            image = qt.QImage(
                int(size[0]), int(size[1]),
                qt.QImage.Format_ARGB32_Premultiplied)
        else:
            # non transparent output
            image = qt.QImage(
                int(size[0]), int(size[1]),
                qt.QImage.Format_RGB32)
            backqcolor.setAlpha(255)

        image.setDotsPerMeterX(int(self.phelpers[0].dpi[0]*m_inch))
        image.setDotsPerMeterY(int(self.phelpers[0].dpi[1]*m_inch))
        if backqcolor.alpha() == 0:
            image.fill(qt.qRgba(0,0,0,0))
        else:
            image.fill(backqcolor.rgb())

        # paint to the image
        painter = qt.QPainter(image)
        painter.setRenderHint(qt.QPainter.Antialiasing, self.aexport.antialias)
        painter.setRenderHint(qt.QPainter.TextAntialiasing, self.aexport.antialias)
        self.phelpers[0].renderToPainter(painter)
        painter.end()

        # write image to disk
        writer = qt.QImageWriter()
        writer.setFormat(fmt.encode('ascii'))
        writer.setFileName(self.filename)

        if fmt == 'png':
            # max compression for PNGs (this number comes from the
            # source code)
            writer.setCompression(100)
            writer.setQuality(0)
        elif fmt == 'tiff':
            # enable LZW compression for TIFFs
            writer.setCompression(1)
        elif fmt == 'jpg':
            # enable optimal JPEG compression using new Qt 5.5 options
            writer.setOptimizedWrite(True)
            writer.setProgressiveScanWrite(True)

        if fmt != 'png':
            writer.setQuality(self.aexport.quality)

        writer.write(image)

class ExportPDFRunnable(ExportRunnable):
    """Runnable task to export a PDF file."""

    def doExport(self):
        """Do the export."""
        printer = qt.QPrinter()
        printer.setResolution(self.aexport.pdfdpi)
        printer.setFullPage(True)
        printer.setColorMode(
            qt.QPrinter.Color if self.aexport.color else qt.QPrinter.GrayScale)
        printer.setOutputFormat(qt.QPrinter.PdfFormat)
        printer.setOutputFileName(self.filename)
        printer.setCreator('Veusz %s' % utils.version())

        def updateSize(ph):
            """Update page size in QPrinter"""
            sizeinchx, sizeinchy = ph.pagesize[0]/ph.dpi[0], ph.pagesize[1]/ph.dpi[1]
            pagesize = qt.QPageSize(
                qt.QSizeF(sizeinchx, sizeinchy), qt.QPageSize.Inch)
            layout = qt.QPageLayout(
                pagesize, qt.QPageLayout.Portrait, qt.QMarginsF())
            printer.setPageLayout(layout)

        updateSize(self.phelpers[0])

        painter = qt.QPainter(printer)
        for i, phelper in enumerate(self.phelpers):
            if i>0:
                updateSize(phelper)
                printer.newPage()
            phelper.renderToPainter(painter)
        painter.end()

class ExportPostscriptRunnable(ExportRunnable):
    """Task to export .ps/.eps files."""

    # whether ghostscript has been searched for
    gs_searched = False
    # its path if it exists
    gs_exe = None
    # map extensions to ghostscript devices
    gs_dev = None

    @classmethod
    def searchGhostscript(klass):
        """Find location of Ghostscript executable."""
        if klass.gs_searched:
            return

        gs_exe = None
        gs = setting.settingdb["external_ghostscript"]
        if gs:
            if os.path.isfile(gs) and os.access(gs, os.X_OK):
                gs_exe = gs
        else:
            if sys.platform == "win32":
                # look for ghostscript as 64 and 32 bit versions
                gs_exe = utils.findOnPath("gswin64c.exe")
                if not gs_exe:
                    gs_exe = utils.findOnPath("gswin32c.exe")
            else:
                # unix tends to call it just gs
                gs_exe = utils.findOnPath("gs")

        klass.gs_dev = dev = {}
        if gs_exe:
            try:
                # check output devices contain
                #  ps2write/eps2write or pswrite/epswrite
                popen = subprocess.Popen(
                    [gs_exe, '-h'],
                    stdout=subprocess.PIPE,
                    universal_newlines=True)
                text = popen.stdout.read()

                if re.search(r'\beps2write\b', text):
                    dev['.eps'] = 'eps2write'
                elif re.search(r'\bepswrite\b', text):
                    dev['.eps'] = 'epswrite'
                if re.search(r'\bps2write\b', text):
                    dev['.ps'] = 'ps2write'
                elif re.search(r'\bpswrite\b', text):
                    dev['.ps'] = 'pswrite'
            except Exception as e:
                pass
            else:
                klass.gs_exe = gs_exe

        klass.gs_searched = True

    def doExport(self):
        """Do the export."""

        self.searchGhostscript()
        if not self.gs_exe:
            raise RuntimeError("Cannot write Postscript without Ghostscript available")

        # write to pdf file first
        ext = os.path.splitext(self.filename)[1].lower()
        tmpfilepdf = "%s.tmp.%i.pdf" % (
            self.filename, random.randint(0,1000000))
        tmpfileps = "%s.tmp.%i%s" % (
            self.filename, random.randint(0,1000000), ext)

        pdfrunnable = ExportPDFRunnable(self.aexport, tmpfilepdf, self.phelpers)
        pdfrunnable.run()

        # run ghostscript to covert from pdf to postscript
        cmd = [
            self.gs_exe,
            '-q', '-dNOCACHE', '-dNOPAUSE', '-dBATCH', '-dSAFER',
            '-sDEVICE=%s' % self.gs_dev[ext],
            '-sOutputFile=%s' % tmpfileps,
            tmpfilepdf
            ]
        try:
            subprocess.check_call(cmd)
        except Exception as e:
            raise RuntimeError("Could not run ghostscript: %s" % str(e))

        if not os.path.isfile(tmpfileps):
            raise RuntimeError("Ghostscript failed to create %s" % tmpfileps)

        os.remove(tmpfilepdf)
        try:
            os.remove(self.filename)
        except OSError:
            pass
        os.rename(tmpfileps, self.filename)

class ExportSVGRunnable(ExportRunnable):
    """Runnable task to export an SVG file."""

    def doExport(self):
        sdpi = self.phelpers[0].dpi
        size = self.phelpers[0].pagesize
        with codecs.open(self.filename, 'w', 'utf-8') as f:
            paintdev = svg_export.SVGPaintDevice(
                f,
                size[0]/sdpi[0], size[1]/sdpi[1],
                writetextastext=self.aexport.svgtextastext,
                dpi=sdpi[1]*svg_dpi_scale,
                scale=svg_dpi_scale)
            self.renderPage(paintdev, self.phelpers[0])

class ExportSelfTestRunnable(ExportRunnable):
    """Runnable task to export a self-test output."""

    def doExport(self):
        sdpi = self.phelpers[0].dpi
        size = self.phelpers[0].pagesize
        with codecs.open(self.filename, 'w', 'utf-8') as f:
            paintdev = selftest_export.SelfTestPaintDevice(
                f,
                size[0]/sdpi[0], size[1]/sdpi[1],
                dpi=sdpi[1])
            self.renderPage(paintdev, self.phelpers[0])

class ExportPICRunnable(ExportRunnable):
    """Runnable task to export Qt PIC output."""

    def doExport(self):
        paintdev = qt.QPicture()
        self.renderPage(paintdev, self.phelpers[0])
        paintdev.save(self.filename)

class ExportEMFRunnable(ExportRunnable):
    """Runnable task to export EMF output."""

    def doExport(self):
        dpi = self.phelpers[0].dpi
        size = self.phelpers[0].pagesize
        paintdev = emf_export.EMFPaintDevice(
            size[0]/dpi[0], size[1]/dpi[1], dpi=dpi[1])
        self.renderPage(paintdev, self.phelpers[0])
        paintdev.paintEngine().saveFile(self.filename)

class AsyncExport(qt.QObject):
    """Asynchronous export.

    Add export tasks with add() and wait with finish().
    """

    @classmethod
    def getFormats(klass):
        """Return list of formats in form of tuples of extension and description."""
        formats = [
            (["bmp"], _("Windows bitmap")),
            (["jpg"], _("Jpeg bitmap")),
            (["pdf"], _("Portable Document Format")),
            (["png"], _("Portable Network Graphics")),
            (["svg"], _("Scalable Vector Graphics")),
            (["tiff"], _("Tagged Image File Format bitmap")),
            (["xpm"], _("X Pixmap")),
        ]

        if hasemf:
            formats.append( (["emf"], _("Windows Enhanced Metafile")) )

        ExportPostscriptRunnable.searchGhostscript()
        if '.eps' in ExportPostscriptRunnable.gs_dev:
            formats.append((["eps"], _("Encapsulated Postscript")))
        if '.ps' in ExportPostscriptRunnable.gs_dev:
            formats.append((["ps"], _("Postscript")))

        formats.sort()
        return formats

    def __init__(self, doc, color=True, bitmapdpi=100,
                 antialias=True, quality=85, backcolor='#ffffff00',
                 pdfdpi=150, svgdpi=96, svgtextastext=False):
        """Initialise export class. Parameters are:
        doc: document to write
        color: use color or try to use monochrome
        bitmapdpi: assume this dpi value when writing images
        antialias: antialias text and lines when writing bitmaps
        quality: compression factor for bitmaps
        backcolor: background color default for bitmaps (default transparent).
        pdfdpi: dpi for pdf and eps files
        svgdpi: dpi for svg files
        svgtextastext: write text in SVG as text, rather than curves
        """

        qt.QObject.__init__(self)
        self.doc = doc
        self.color = color
        self.bitmapdpi = bitmapdpi
        self.antialias = antialias
        self.quality = quality
        self.backcolor = backcolor
        self.pdfdpi = pdfdpi
        self.svgdpi = svgdpi
        self.svgtextastext = svgtextastext

        self.backqcolor = self.doc.evaluate.colors.get(self.backcolor)

        # any exceptions in runnables should set this to be reported
        # in main thread for UI
        self.exception = None

        # pool that export threads use to execute
        self.pool = qt.QThreadPool(self)
        self.pool.setMaxThreadCount(
            max(setting.settingdb['plot_numthreads'], 1)
        )

    def finish(self):
        self.pool.waitForDone()

        if self.exception is not None:
            exception = self.exception
            self.exception = None
            raise exception

    def haveDone(self):
        """Have all the threads finished?

        Note: call finish afterwards for cleanup
        """
        return self.pool.waitForDone(0)

    def getDPI(self, ext):
        """Get DPI to use for filename extension."""

        if ext in {'.pdf', '.eps', '.ps'}:
            return (self.pdfdpi, self.pdfdpi)

        elif ext in {'.png', '.jpg', '.jpeg', '.bmp', '.tiff', '.xpm'}:
            return (self.bitmapdpi, self.bitmapdpi)

        elif ext == '.svg':
            dpi = self.svgdpi / svg_dpi_scale
            return (dpi, dpi)

        elif ext == '.selftest':
            return (90, 90)

        elif ext == '.pic':
            pic = qt.QPicture()
            return (pic.logicalDpiX(), pic.logicalDpiY())

        elif ext == '.emf':
            return (90, 90)

        else:
            raise RuntimeError('Unknown export file type')

    def add(self, filename, pages):
        """Add export to list to be processed.

        filename: output filename
        pages: list of pages (0-indexed) to be added.

        sync: if True, then do not execute in different thread."""

        ext = os.path.splitext(filename)[1].lower()
        dpi = self.getDPI(ext)

        # render each page to a PaintHelper
        phelpers = []
        for page in pages:
            size = self.doc.pageSize(page, dpi=dpi, integer=False)
            phelper = painthelper.PaintHelper(self.doc, size, dpi=dpi)
            self.doc.paintTo(phelper, page)
            phelpers.append(phelper)

        # single page only formats
        if len(phelpers) != 1 and ext not in ('.ps', '.pdf'):
            raise RuntimeError('Only single page allowed for format')

        # make a runnable task for the right file type
        runnable = {
            '.png': ExportBitmapRunnable,
            '.jpg': ExportBitmapRunnable,
            '.jpeg': ExportBitmapRunnable,
            '.bmp': ExportBitmapRunnable,
            '.tiff': ExportBitmapRunnable,
            '.xpm': ExportBitmapRunnable,

            '.pdf': ExportPDFRunnable,
            '.ps': ExportPostscriptRunnable,
            '.eps': ExportPostscriptRunnable,

            '.svg': ExportSVGRunnable,
            '.selftest': ExportSelfTestRunnable,
            '.pic': ExportPICRunnable,
            '.emf': ExportEMFRunnable,
        }[ext](self, filename, phelpers)

        self.pool.start(runnable)

def printPages(doc, printer, pages, scaling=1., antialias=False, setsizes=False):
    """Print onto printing device.
    Returns list of page sizes
    setsizes: Set page size on printer to page sizes
    """

    if not pages:
        return

    dpi = (printer.logicalDpiX(), printer.logicalDpiY())

    def getUpdateSize(page):
        size = doc.pageSize(page, dpi=dpi, integer=False)
        if setsizes:
            # update paper size on printer
            sizeinchx, sizeinchy = size[0]/dpi[0], size[1]/dpi[1]
            pagesize = qt.QPageSize(
                qt.QSizeF(sizeinchx, sizeinchy), qt.QPageSize.Inch)
            layout = qt.QPageLayout(
                pagesize, qt.QPageLayout.Portrait, qt.QMarginsF())
            printer.setPageLayout(layout)
        return size

    size = getUpdateSize(pages[0])

    painter = painthelper.DirectPainter(printer)
    if antialias:
        painter.setRenderHint(qt.QPainter.Antialiasing, True)
        painter.setRenderHint(qt.QPainter.TextAntialiasing, True)

    # This all assumes that only pages can go into the root widget
    visible = set(doc.getVisiblePages())
    filtpages = [page for page in pages if page in visible]

    for count, page in enumerate(filtpages):
        psize = doc.pageSize(page, dpi=dpi, integer=False, scaling=scaling)
        phelper = painthelper.PaintHelper(doc, psize, dpi=dpi, scaling=scaling)
        doc.paintTo(phelper, page)
        phelper.renderToPainter(painter)

        # start new pages between each page
        if count < len(filtpages)-1:
            # set page size before newPage!
            size = getUpdateSize(pages[count+1])
            printer.newPage()

    painter.end()

def printDialog(parentwindow, document, filename=None):
    """Open a print dialog and print document."""

    if not document.getVisiblePages():
        qt.QMessageBox.warning(
            parentwindow, _("Error - Veusz"), _("No pages to print"))
        return

    prnt = qt.QPrinter(qt.QPrinter.HighResolution)
    prnt.setColorMode(qt.QPrinter.Color)
    prnt.setCreator(_('Veusz %s') % utils.version())
    if filename:
        prnt.setDocName(filename)

    dialog = qt.QPrintDialog(prnt, parentwindow)
    dialog.setMinMax(1, document.getNumberPages())
    if dialog.exec_():
        # get page range
        if dialog.printRange() == qt.QAbstractPrintDialog.PageRange:
            # page range
            minval, maxval = dialog.fromPage(), dialog.toPage()
        else:
            # all pages
            minval, maxval = 1, document.getNumberPages()

        # pages are relative to zero
        minval -= 1
        maxval -= 1

        # reverse or forward order
        if prnt.pageOrder() == qt.QPrinter.FirstPageFirst:
            pages = list(range(minval, maxval+1))
        else:
            pages = list(range(maxval, minval-1, -1))

        # if more copies are requested
        pages *= prnt.copyCount()

        # do the printing
        printPages(document, prnt, pages)



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

from .field import *
from .datasetplugin import *
from .importplugin import *
from .toolsplugin import *
from .votable import *

# backward compatibility
ImportDataset1D = Dataset1D
ImportDataset2D = Dataset2D
ImportDatasetText = DatasetText
ImportField = Field
ImportFieldCheck = FieldBool
ImportFieldText = FieldText
ImportFieldFloat = FieldFloat
ImportFieldInt = FieldInt
ImportFieldCombo = FieldCombo



# graph widget for containing other sorts of widget

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import textwrap

from .. import qtall as qt
from .. import setting
from .. import utils
from .. import document

from . import widget
from . import controlgraph

def _(text, disambiguation=None, context='Graph'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class Graph(widget.Widget):
    """Graph for containing other sorts of widgets"""

    typename='graph'
    allowusercreation = True
    description = _('Base graph')

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        widget.Widget.addSettings(s)

        s.add( setting.Distance(
            'leftMargin',
            '1.7cm',
            descr=_('Distance from left of graph to edge'),
            usertext=_('Left margin'),
            formatting=True,
        ) )
        s.add( setting.Distance(
            'rightMargin',
            '0.2cm',
            descr=_('Distance from right of graph to edge'),
            usertext=_('Right margin'),
            formatting=True,
        ) )
        s.add( setting.Distance(
            'topMargin',
            '0.2cm',
            descr=_('Distance from top of graph to edge'),
            usertext=_('Top margin'),
            formatting=True,
        ) )
        s.add( setting.Distance(
            'bottomMargin',
            '1.7cm',
            descr=_('Distance from bottom of graph to edge'),
            usertext=_('Bottom margin'),
            formatting=True,
        ) )
        s.add( setting.FloatOrAuto(
            'aspect',
            'Auto',
            descr=_('Fix aspect ratio of graph to this value'),
            usertext=_('Aspect ratio'),
            minval = 0.01,
            maxval = 100.,
            formatting=True,
        ) )

        s.add( setting.Notes(
            'notes', '',
            descr=_('User-defined notes'),
            usertext=_('Notes')
        ) )

        s.add(
            setting.GraphBrush(
                'Background',
                descr = _('Background plot fill'),
                usertext=_('Background')),
            pixmap='settings_bgfill',
        )
        s.add(
            setting.Line(
                'Border',
                descr = _('Graph border line'),
                usertext=_('Border')),
            pixmap='settings_border',
        )

    @classmethod
    def allowedParentTypes(klass):
        from . import page, grid
        return (page.Page, grid.Grid)

    @property
    def userdescription(self):
        """Return user-friendly description."""
        return textwrap.fill(self.settings.notes, 60)

    def addDefaultSubWidgets(self):
        """Add axes automatically."""

        from . import axis
        if self.parent.getChild('x') is None:
            ax = axis.Axis(self, name='x')
            ax.linkToStylesheet()
        if self.parent.getChild('y') is None:
            ay = axis.Axis(self, name='y')
            ay.linkToStylesheet()

    def getAxesDict(self, axesnames, ignoremissing=False):
        """Get the axes for widgets to plot against.
        axesnames is a list/set of names to find.
        Returns a dict of objects
        """

        axes = {}
        # recursively go back up the tree to find axes
        w = self
        while w is not None and len(axes) < len(axesnames):
            for c in w.children:
                name = c.name
                if ( name in axesnames and name not in axes and
                     c.isaxis ):
                    axes[name] = c
            w = w.parent

        # didn't find everything...
        if w is None and not ignoremissing:
            for name in axesnames:
                if name not in axes:
                    axes[name] = None

        # return list of found axes
        return axes

    def getAxes(self, axesnames):
        """Return a list of axes widgets given a list of names."""
        ad = self.getAxesDict(axesnames)
        return [ad[n] for n in axesnames]

    def adjustBoundsForAspect(self, bounds):
        s = self.settings
        if s.aspect != 'Auto':
            saspect = s.aspect
            width = bounds[2]-bounds[0]
            height = bounds[3]-bounds[1]
            gaspect = width/height

            bounds = list(bounds)
            if saspect > gaspect:
                # want a graph which is wider than the current size
                # => add space to top/bottom
                newheight = width / saspect
                delta = (height-newheight) / 2
                bounds[1] += delta
                bounds[3] -= delta
            else:
                # want a graph which is narrower than the current size
                # => add space to left/right
                newwidth = height * saspect
                delta = (width-newwidth) / 2
                bounds[0] += delta
                bounds[2] -= delta

        return bounds

    def getMargins(self, painthelper):
        """Use settings to compute margins."""
        s = self.settings
        return (
            s.get('leftMargin').convert(painthelper),
            s.get('topMargin').convert(painthelper),
            s.get('rightMargin').convert(painthelper),
            s.get('bottomMargin').convert(painthelper),
        )

    def draw(self, parentposn, painthelper, outerbounds = None):
        '''Update the margins before drawing.'''

        # yuck, avoid circular imports
        from . import axisbroken

        s = self.settings

        bounds = self.computeBounds(parentposn, painthelper)
        maxbounds = self.computeBounds(
            parentposn, painthelper, withmargin=False)

        # do no painting if hidden
        if s.hide:
            return bounds

        # controls for adjusting graph margins
        painter = painthelper.painter(self, bounds)
        painthelper.setControlGraph(self, [
            controlgraph.ControlMarginBox(self, bounds, maxbounds, painthelper)
        ])

        bounds = self.adjustBoundsForAspect(bounds)

        with painter:
            # set graph rectangle attributes
            path = qt.QPainterPath()
            path.addRect(qt.QRectF(
                qt.QPointF(bounds[0], bounds[1]),
                qt.QPointF(bounds[2], bounds[3]) ))
            utils.brushExtFillPath(
                painter, s.Background, path,
                stroke=s.Border.makeQPenWHide(painter))

            # debugging positions (uncomment)
            # painter.drawRect( qt.QRectF(
            #         qt.QPointF(parentposn[0], parentposn[1]),
            #         qt.QPointF(parentposn[2], parentposn[3]) ))

            # if outerbounds:
            #     painter.drawRect( qt.QRectF(
            #             qt.QPointF(outerbounds[0], outerbounds[1]),
            #             qt.QPointF(outerbounds[2], outerbounds[3]) ))

        # child drawing algorithm is a bit complex due to axes
        # being shared between graphs and broken axes

        # this is a map of axis names to plot to axis widgets
        axestodraw = {}
        # axes widgets for each plotter (precalculated by Page)
        axesofwidget = painthelper.plotteraxismap
        for c in self.children:
            try:
                for a in axesofwidget[c]:
                    axestodraw[a.name] = a
            except (KeyError, AttributeError):
                if c.isaxis:
                    axestodraw[c.name] = c

        # grid lines are normally plotted before other child widgets
        axisdrawlist = sorted(axestodraw.items(), reverse=True)
        for aname, awidget in axisdrawlist:
            awidget.updateAxisLocation(bounds)
            awidget.computePlottedRange()
            awidget.drawGrid(
                bounds, painthelper, outerbounds=outerbounds, ontop=False)

        # broken axis handling
        brokenaxes = set()
        for axis in axestodraw.values():
            if isinstance(axis, axisbroken.AxisBroken):
                brokenaxes.add(axis)

        # don't duplicate drawing axes
        axesdrawn = set()

        # reset counter and compute automatic colors
        painthelper.autoplottercount = 0
        for c in self.children:
            c.setupAutoColor(painter)

        # do normal drawing of children
        # iterate over children in reverse order
        for c in reversed(self.children):

            if c.isaxis:
                axesdrawn.add(c)

            axes = axesofwidget.get(c, None)
            if axes is not None and any((a in brokenaxes for a in axes)):
                # handle broken axes
                childbrokenaxes = sorted(
                    [(a.name, a) for a in axes if a in brokenaxes],
                    key=lambda x: x[0])
                def iteratebrokenaxes(b):
                    """Recursively iterate over each broken axis and redraw
                    child for each.
                    We might have more than one broken axis per child, so
                    hence this rather strange iteration.
                    """
                    ax = b[0][1]
                    for i in range(ax.breakvnum):
                        ax.switchBreak(i, bounds)
                        if len(b) == 1:
                            c.draw(bounds, painthelper, outerbounds=outerbounds)
                        else:
                            iteratebrokenaxes(b[1:])
                    ax.switchBreak(None, bounds)
                iteratebrokenaxes(childbrokenaxes)

            else:

                # standard non broken axis drawing
                c.draw(bounds, painthelper, outerbounds=outerbounds)

        # then for grid lines on top
        axiswidgets = [axis for name, axis in axisdrawlist]
        for awidget in axiswidgets:
            awidget.drawGrid(
                bounds, painthelper, outerbounds=outerbounds, ontop=True)
            awidget.drawAutoMirror(bounds, painthelper, axiswidgets)

        # draw remaining axes
        for awidget in axiswidgets:
            if awidget not in axesdrawn:
                awidget.draw(bounds, painthelper, outerbounds=outerbounds)

        return bounds

    def updateControlItem(self, cgi):
        """Graph resized or moved - call helper routine to move self."""
        cgi.setWidgetMargins()

# allow users to make Graph objects
document.thefactory.register(Graph)



# axisticks.py
# algorithm to work out what tick-marks to put on an axis

#    Copyright (C) 2003 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import math
import numpy as N

from .. import utils

"""Algorithms for working with axis ticks.

These algorithms were designed by me (Jeremy Sanders), so there
may well be bugs. Please report them.

The idea is to try to achieve a set number of major and minor ticks
by looking though a list of allowable interval values (after taking
account of what power of 10 the coordinates are in).
"""

class AxisTicksBase:
    """Base class of axis ticks classes."""

    def __init__( self, minval, maxval, numticks, numminorticks,
                  logaxis = False, prefermore = True,
                  extendmin = False, extendmax = False,
                  forceinterval = None ):
        """Initialise the class.

        minval and maxval are the range of the data to be plotted
        numticks number of major ticks to aim for
        logaxis: axis logarithmic?
        prefermore: prefer more ticks rather than fewer
        extendbounds: extend minval and maxval to nearest tick if okay
        forceinterval: force interval to one given (if allowed). interval
         is tuple as returned in self.interval after calling getTicks()
        """

        # clip to sensible range
        self.minval = max(min(minval, 1e100), -1e100)
        self.maxval = max(min(maxval, 1e100), -1e100)

        # tick parameters
        self.numticks = numticks
        self.numminorticks = numminorticks
        self.logaxis = logaxis
        self.prefermore = prefermore
        self.extendmin = extendmin
        self.extendmax = extendmax
        self.forceinterval = forceinterval

    def getTicks( self ):
        """Calculate and return the position of the major ticks.

        Results are returned as attributes of this object in
        interval, minval, maxval, tickvals, minorticks, autoformat
        """

class AxisTicks(AxisTicksBase):
    """Class to work out at what values axis major ticks should appear."""

    # the allowed values we allow ticks to increase by
    # first values are the major tick intervals, followed by a list
    # of allowed minors
    allowed_minorintervals_linear = { 
        1.:  (0.1, 0.2, 0.5),
        2.:  (0.2, 0.5, 1.),
        5.:  (0.5, 1., 2.5),
        2.5: (0.5,)
    }
    # just get the allowable majors
    allowed_intervals_linear = sorted(allowed_minorintervals_linear)

    # the allowed values we can increase by in log space
    # by default we increase by 10^3
    # if the first value is chosen we can use the "special" log minor ticks
    allowed_intervals_log = (1., 3., 6., 9., 12., 15., 19.)

    # positions we're allowed to put minor intervals
    allowed_minorintervals_log = (1., 3., 6., 9., 12., 15., 19.)

    # how much we should allow axes to extend to a tick
    max_extend_factor = 0.15

    def _calcTickValues( self, minval, maxval, delta ):
        """Compute the tick values, given minval, maxval and delta."""

        startmult = int( math.ceil( minval / delta ) )
        stopmult = int( math.floor( maxval / delta ) )

        return N.arange(startmult, stopmult+1) * delta

    def _tickNums(self, minval, maxval, delta):
        """Calculate number of ticks between minval and maxval with delta."""

        startmult = int( math.ceil( minval / delta ) )
        stopmult = int( math.floor( maxval / delta ) )

        return (stopmult-startmult)+1

    def _calcNoTicks( self, interval, logdelta ):
        """Return the number of ticks with spacing interval*10^logdelta.

        Returns a tuple (noticks, minval, maxval).
        """

        # store these for modification (if we extend bounds)
        minval = self.minval
        maxval = self.maxval

        # calculate tick spacing and maximum extension factor
        delta = interval * (10**logdelta)
        maxextend = (maxval - minval) * AxisTicks.max_extend_factor

        # should we try to extend to nearest interval*10^logdelta?
        if self.extendmin:
            # extend minval if possible
            if math.fabs( math.modf( minval / delta )[0] ) > 1e-8:
                d = minval - ( math.floor( minval / delta ) * delta )
                if d <= maxextend:
                    minval -= d

        if self.extendmax:
            # extend maxval if possible
            if math.fabs( math.modf( maxval / delta)[0] ) > 1e-8:
                d = ( (math.floor(maxval / delta)+1.) * delta) - maxval
                if d <= maxextend:
                    maxval += d

        numticks = self._tickNums(minval, maxval, delta)
        return (numticks, minval, maxval)

    def _calcLinearMinorTickValues(self, minval, maxval, interval, logstep,
                                   allowedintervals):
        """Get the best values for minor ticks on a linear axis

        Algorithm tries to look for best match to nominorticks
        Pass routine major ticks from minval to maxval with steps of
        interval*(10**logstep)
        """

        # iterate over allowed minor intervals
        best = -1
        best_numticks = -1
        best_delta = 1000000
        mult = 10.**logstep

        # iterate over allowed minor intervals
        for minint in allowedintervals:

            numticks = self._tickNums(minval, maxval, minint*mult)
            d = abs( self.numminorticks - numticks )

            # if this is a better match to the number of ticks
            # we want, choose this
            if ((d < best_delta ) or
                (d == best_delta and
                (self.prefermore and numticks > best_numticks) or
                (not self.prefermore and numticks < best_numticks)) ):

                best = minint
                best_delta = d
                best_numticks = numticks

        # use best value to return tick values
        return self._calcTickValues(minval, maxval, best*mult)

    def _calcLogMinorTickValues( self, minval, maxval ):
        """Calculate minor tick values with a log scale."""

        # this is a scale going e.g. 1,2,3,...8,9,10,20,30...90,100,200...

        # round down to nearest power of 10 for each
        alpha = int( math.floor( N.log10(minval) ) )
        beta = int( math.floor( N.log10(maxval) ) )

        ticks = []
        # iterate over range in log space
        for i in range(alpha, beta+1):
            power = 10.**i
            # add ticks for values in correct range
            for j in range(2, 10):
                v = power*j
                # blah log conversions mean we have to use 'fuzzy logic'
                if ( math.fabs(v - minval)/v < 1e-6 or v > minval ) and \
                   ( math.fabs(v - maxval)/v < 1e-6 or v < maxval ) :
                    ticks.append(v)

        return N.array( ticks )

    def _selectBestTickFromSelection(self, selection):
        """Choose best tick from selection given."""
        # we now try to find the best matching value
        minabsdelta = 1e99
        mindelta = 1e99
        bestsel = ()

        # find the best set of tick labels
        for s in selection:
            # difference between what we want and what we have
            delta = s[0] - self.numticks
            absdelta = abs(delta)

            # if it matches better choose this
            if absdelta < minabsdelta:
                minabsdelta = absdelta
                mindelta = delta
                bestsel = s

            # if we find two closest matching label sets, we
            # test whether we prefer too few to too many labels
            if absdelta == minabsdelta:
                if (self.prefermore and (delta > mindelta)) or \
                       (not self.prefermore and (delta < mindelta)):
                    minabsdelta = absdelta
                    mindelta = delta
                    bestsel = s

        return bestsel

    def _getBestTickSelection(self, allowed_intervals):
        """Go through allowed tick intervals and find one best matching
        requested parameters."""

        # work out range and log range
        therange = self.maxval - self.minval
        intlogrange = int( N.log10( therange ) )

        # we step variable to move through log space to find best ticks
        logstep = intlogrange + 1

        # we iterate down in log spacing, until we have more than twice
        # the number of ticks requested.
        # Maybe a better algorithm is required
        selection = []

        # keep track of largest number of ticks calculated
        largestno = 0

        while True:
            for interval in allowed_intervals:
                no, minval, maxval = self._calcNoTicks( interval, logstep )
                selection.append( (no, interval, logstep, minval, maxval ) )

                largestno = max(largestno, no)

            if largestno > self.numticks*2:
                break

            logstep -= 1

            # necessary as we don't want 10**x on axis if |x|<1
            # :-(
            if logstep < 0 and self.logaxis:
                break

        return selection

    def _tickSelector(self, allowed_intervals):
        """With minval and maxval find best tick positions."""

        if self.forceinterval is None:
            # get selection of closely matching ticks
            selection = self._getBestTickSelection(allowed_intervals)

            # now we have the best, we work out the ticks and return
            bestsel = self._selectBestTickFromSelection(selection)
            dummy, interval, loginterval, minval, maxval = bestsel
        else:
            # forced specific interval requested
            interval, loginterval = self.forceinterval
            no, minval, maxval = self._calcNoTicks(interval, loginterval)

        # calculate the positions of the ticks from parameters
        tickdelta = interval * 10.**loginterval
        ticks = self._calcTickValues( minval, maxval, tickdelta )

        return (minval, maxval, ticks, interval, loginterval)

    def getTicks(self):
        """Calculate and return the position of the major ticks.
        """

        if self.logaxis:
            # which intervals we'll accept for major ticks
            intervals = AxisTicks.allowed_intervals_log

            # transform range into log space
            self.minval = N.log10( self.minval )
            self.maxval = N.log10( self.maxval )
        else:
            # which linear intervals we'll allow
            intervals = AxisTicks.allowed_intervals_linear

        # avoid breakage if range is zero
        if abs(self.minval - self.maxval) < 1e-99:
            self.maxval = self.minval + 1.

            # for large numbers this may be the case
            if self.maxval == self.minval:
                self.minval, self.maxval = 0., 1.

        minval, maxval, tickvals, interval, loginterval = self._tickSelector(
            intervals )

        # work out the most appropriate minor tick intervals
        if not self.logaxis:
            # just plain minor ticks
            # try to achieve no of minors close to value requested

            minorticks = self._calcLinearMinorTickValues(
                minval, maxval, interval, loginterval,
                AxisTicks.allowed_minorintervals_linear[interval]
            )

        else:
            # log axis
            if interval == 1.:
                # calculate minor ticks
                # here we use 'conventional' minor log tick spacing
                # e.g. 0.9, 1, 2, .., 8, 9, 10, 20, 30 ...
                minorticks = self._calcLogMinorTickValues(
                    10.**minval, 10.**maxval)

                # Here we test whether more log major tick values are needed...
                # often we might only have one tick value, and so we add 2, then 5
                # this is a bit of a hack: better ideas please!!

                if len(tickvals) < 2:
                    # get lower power of 10
                    low10 = int( math.floor(minval) )

                    # could use numpy here
                    for i in (2., 5., 20., 50.):
                        n = low10 + math.log10(i)
                        if n >= minval and n <= maxval:
                            tickvals = N.concatenate( (tickvals, N.array([n]) ))

            else:
                # if we increase by more than one power of 10 on the
                # axis, we can't do the above, so we do linear ticks
                # in log space
                # aim is to choose powers of 3 for majors and minors
                # to make it easy to read the axis. comments?

                minorticks = self._calcLinearMinorTickValues(
                    minval, maxval, interval, loginterval,
                    AxisTicks.allowed_minorintervals_log)
                minorticks = 10.**minorticks

            # transform normal ticks back to real space
            minval = 10.**minval
            maxval = 10.**maxval
            tickvals = 10.**tickvals

        self.interval = (interval, loginterval)
        self.minorticks = minorticks
        self.minval = minval
        self.maxval = maxval
        self.tickvals = tickvals
        self.autoformat = '%Vg'

class DateTicks(AxisTicksBase):
    """For formatting dates. We want something that chooses appropriate
    intervals
    So we want to choose most apropriate interval depending on number of
    ticks requested
    """

    # possible intervals for a time/date axis
    # tuples of ((y, m, d, h, m, s, msec), autoformat)
    intervals = (
        ((200, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((100, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((50, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((20, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((10, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((5, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((2, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((1, 0, 0, 0, 0, 0, 0), '%VDY'),
        ((0, 6, 0, 0, 0, 0, 0), '%VDY-%VDm'),
        ((0, 4, 0, 0, 0, 0, 0), '%VDY-%VDm'),
        ((0, 3, 0, 0, 0, 0, 0), '%VDY-%VDm'),
        ((0, 2, 0, 0, 0, 0, 0), '%VDY-%VDm'),
        ((0, 1, 0, 0, 0, 0, 0), '%VDY-%VDm'),
        ((0, 0, 28, 0, 0, 0, 0), '%VDY-%VDm-%VDd'),
        ((0, 0, 14, 0, 0, 0, 0), '%VDY-%VDm-%VDd'),
        ((0, 0, 7, 0, 0, 0, 0), '%VDY-%VDm-%VDd'),
        ((0, 0, 2, 0, 0, 0, 0), '%VDY-%VDm-%VDd'),
        ((0, 0, 1, 0, 0, 0, 0), '%VDY-%VDm-%VDd'),
        ((0, 0, 0, 12, 0, 0, 0), '%VDY-%VDm-%VDd\\\\%VDH:%VDM'),
        ((0, 0, 0, 6, 0, 0, 0), '%VDY-%VDm-%VDd\\\\%VDH:%VDM'),
        ((0, 0, 0, 4, 0, 0, 0), '%VDY-%VDm-%VDd\\\\%VDH:%VDM'),
        ((0, 0, 0, 3, 0, 0, 0), '%VDY-%VDm-%VDd\\\\%VDH:%VDM'),
        ((0, 0, 0, 2, 0, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 1, 0, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 30, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 15, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 10, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 5, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 2, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 1, 0, 0), '%VDH:%VDM'),
        ((0, 0, 0, 0, 0, 30, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 15, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 10, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 5, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 2, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 1, 0), '%VDH:%VDM:%VDS'),
        ((0, 0, 0, 0, 0, 0, 500000), '%VDH:%VDM:%VDVS'),
        ((0, 0, 0, 0, 0, 0, 200000), '%VDVS'),
        ((0, 0, 0, 0, 0, 0, 100000), '%VDVS'),
        ((0, 0, 0, 0, 0, 0, 50000), '%VDVS'),
        ((0, 0, 0, 0, 0, 0, 10000), '%VDVS'),
    )

    intervals_sec = N.array([
        (
            ms*1e-6+s+mi*60+hr*60*60+dy*24*60*60 +
            mn*(365/12.)*24*60*60 +
            yr*365*24*60*60
        )
        for (yr, mn, dy, hr, mi, s, ms), fmt in intervals]
    )

    def bestTickFinder(self, minval, maxval, numticks, extendmin, extendmax,
                       intervals, intervals_sec):
        """Try to find best choice of numticks ticks between minval and maxval
        intervals is an array similar to self.intervals
        intervals_sec is an array similar to self.intervals_sec

        Returns a tuple (minval, maxval, estimatedsize, ticks, textformat)"""

        delta = maxval - minval

        # iterate over different intervals and find one closest to what we want
        estimated = delta / intervals_sec

        tick1 = max(estimated.searchsorted(numticks)-1, 0)
        tick2 = min(tick1+1, len(estimated)-1)

        del1 = abs(estimated[tick1] - numticks)
        del2 = abs(estimated[tick2] - numticks)

        if del1 < del2:
            best = tick1
        else:
            best = tick2
        besttt, fmt = intervals[best]

        mindate = utils.floatToDateTime(minval)
        maxdate = utils.floatToDateTime(maxval)

        # round min and max to nearest
        minround = utils.tupleToDateTime(utils.roundDownToTimeTuple(mindate, besttt))
        maxround = utils.tupleToDateTime(utils.roundDownToTimeTuple(maxdate, besttt))

        if minround == mindate:
            mintick = minround
        else:
            # rounded down, so move on to next tick
            mintick = utils.addTimeTupleToDateTime(minround, besttt)
        maxtick = maxround

        # extend bounds if requested
        deltamin = utils.datetimeToFloat(mindate)-utils.datetimeToFloat(mintick)
        if extendmin and (deltamin != 0. and deltamin < delta*0.15):
            mindate = utils.addTimeTupleToDateTime(
                minround, [-x for x in besttt])
            mintick = mindate
        deltamax = utils.datetimeToFloat(maxdate)-utils.datetimeToFloat(maxtick)
        if extendmax and (deltamax != 0. and deltamax < delta*0.15):
            maxdate = utils.addTimeTupleToDateTime(maxtick, besttt)
            maxtick = maxdate

        # make ticks
        ticks = []
        dt = mintick
        while dt <= maxtick:
            ticks.append( utils.datetimeToFloat(dt))
            dt = utils.addTimeTupleToDateTime(dt, besttt)

        return (
            utils.datetimeToFloat(mindate),
            utils.datetimeToFloat(maxdate),
            intervals_sec[best],
            N.array(ticks), fmt
        )

    def filterIntervals(self, estint):
        """Filter intervals and intervals_sec to be
        multiples of estint seconds."""
        intervals = []
        intervals_sec = []
        for i, inter in enumerate(self.intervals_sec):
            ratio = estint / inter
            if abs(ratio-int(ratio)) < ratio*.01:
                intervals.append(self.intervals[i])
                intervals_sec.append(inter)
        return intervals, N.array(intervals_sec)

    def getTicks(self):
        """Calculate and return the position of the major ticks.
        """

        # find minor ticks
        mindate, maxdate, est, ticks, fmt = self.bestTickFinder(
            self.minval, self.maxval, self.numticks,
            self.extendmin, self.extendmax,
            self.intervals, self.intervals_sec)

        # try to make minor ticks divide evenly into major ticks
        intervals, intervals_sec = self.filterIntervals(est)
        # get minor ticks
        ig, ig, ig, minorticks, ig = self.bestTickFinder(
            mindate, maxdate, self.numminorticks, False, False,
            intervals, intervals_sec)

        self.interval = (intervals, intervals_sec)
        self.minval = mindate
        self.maxval = maxdate
        self.minorticks = minorticks
        self.tickvals = ticks
        self.autoformat = fmt



#    Copyright (C) 2008 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""Plotting a line with arrowheads or labels."""

import math
import itertools
import numpy as N

from .. import qtall as qt
from .. import setting
from .. import document
from .. import utils

from . import controlgraph
from . import plotters

def _(text, disambiguation=None, context='Line'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class Line(plotters.FreePlotter):
    """A line on the plot/graph."""
    typename='line'
    description=_('Line or arrow')
    allowusercreation = True

    @staticmethod
    def showOrHideSetn(v):
        """Whether to show or hide length/angle or position."""
        a = ('length', 'angle')
        b = ('xPos2', 'yPos2')
        return (a,b) if v=='length-angle' else (b,a)

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        plotters.FreePlotter.addSettings(s)

        s.add( setting.ChoiceSwitch(
            'mode',
            ('length-angle', 'point-to-point'),
            'length-angle',
            descr=_(
                'Provide line position and length,angle '
                'or first and second points'),
            usertext=_('Mode'),
            formatting=False,
            showfn=klass.showOrHideSetn), 0)

        s.add( setting.DatasetExtended(
            'length',
            [0.2],
            descr=_('List of fractional lengths, dataset or expression'),
            usertext=_('Lengths'),
            formatting=False), 4 )
        s.add( setting.DatasetExtended(
            'angle',
            [0.],
            descr=_(
                'List of angle of lines, dataset or expression (degrees)'),
            usertext=_('Angles'),
            formatting=False), 5 )
        s.add( setting.DatasetExtended(
            'xPos2',
            [1.],
            descr=_(
                'List of fractional X coordinates, dataset or '
                'expression for point 2'),
            usertext=_('X positions 2'),
            formatting=False), 6 )
        s.add( setting.DatasetExtended(
            'yPos2',
            [1.],
            descr=_(
                'List of fractional Y coordinates, dataset or '
                'expression for point 2'),
            usertext=_('Y positions 2'),
            formatting=False), 7 )

        s.add( setting.Bool(
            'clip', False,
            descr=_('Clip line to its container'),
            usertext=_('Clip'),
            formatting=True), 0 )

        s.add( setting.Line(
            'Line',
            descr=_('Line style'),
            usertext=_('Line')),
            pixmap='settings_plotline' )
        s.add( setting.ArrowFill(
            'Fill',
            descr=_('Arrow fill settings'),
            usertext=_('Arrow fill')),
            pixmap='settings_plotmarkerfill' )

        s.add( setting.DistancePt(
            'arrowSize', '5pt',
            descr=_('Size of arrow to plot'),
            usertext=_('Arrow size'), formatting=True), 0)
        s.add( setting.Arrow(
            'arrowright', 'none',
            descr=_('Arrow to plot on right side'),
            usertext=_('Arrow right'), formatting=True), 0)
        s.add( setting.Arrow(
            'arrowleft', 'none',
            descr=_('Arrow to plot on left side'),
            usertext=_('Arrow left'), formatting=True), 0)

    def _computeLinesLengthAngle(self, posn, lengthscaling):
        """Return set of lines to plot for length-angle."""

        s = self.settings
        d = self.document

        # translate coordinates from axes or relative values
        xpos, ypos = self._getPlotterCoords(posn)
        # get lengths and angles of lines
        length = s.get('length').getFloatArray(d)
        angle = s.get('angle').getFloatArray(d)
        if xpos is None or ypos is None or length is None or angle is None:
            return None
        length *= lengthscaling

        maxlen = max( len(xpos), len(ypos), len(length), len(angle) )
        if maxlen > 1:
            if len(xpos) == 1: xpos = itertools.cycle(xpos)
            if len(ypos) == 1: ypos = itertools.cycle(ypos)
            if len(length) == 1: length = itertools.cycle(length)
            if len(angle) == 1: angle = itertools.cycle(angle)

        out = []
        for v in zip(xpos, ypos, length, angle):
            # skip lines which have nans
            if N.all( N.isfinite(v) ):
                out.append(v)
        return out

    def _computeLinesPointToPoint(self, posn):
        """Return set of lines for point to point."""

        # translate coordinates from axes or relative values
        xpos, ypos = self._getPlotterCoords(posn)
        xpos2, ypos2 = self._getPlotterCoords(posn, xsetting='xPos2', ysetting='yPos2')
        if xpos is None or ypos is None or xpos2 is None or ypos2 is None:
            return None

        maxlen = max( len(xpos), len(ypos), len(xpos2), len(ypos2) )
        if maxlen > 1:
            if len(xpos) == 1: xpos = itertools.cycle(xpos)
            if len(ypos) == 1: ypos = itertools.cycle(ypos)
            if len(xpos2) == 1: xpos2 = itertools.cycle(xpos2)
            if len(ypos2) == 1: ypos2 = itertools.cycle(ypos2)

        out = []
        for v in zip(xpos, ypos, xpos2, ypos2):
            # skip nans again
            if N.all( N.isfinite(v) ):
                length = math.sqrt( (v[0]-v[2])**2 + (v[1]-v[3])**2 )
                angle = math.atan2( v[3]-v[1], v[2]-v[0] ) / math.pi * 180.
                out.append( (v[0], v[1], length, angle) )
        return out

    def draw(self, posn, phelper, outerbounds = None):
        """Plot the key on a plotter."""

        s = self.settings
        d = self.document
        if s.hide:
            return

        # if a dataset is used, we can't use control items
        isnotdataset = (
            not s.get('xPos').isDataset(d) and
            not s.get('yPos').isDataset(d) )

        if s.mode == 'length-angle':
            isnotdataset = (
                isnotdataset and
                not s.get('length').isDataset(d) and
                not s.get('angle').isDataset(d) )
        else:
            isnotdataset = (
                isnotdataset and
                not s.get('xPos2').isDataset(d) and
                not s.get('yPos2').isDataset(d) )

        # now do the drawing
        clip = None
        if s.clip:
            clip = qt.QRectF(
                qt.QPointF(posn[0], posn[1]),
                qt.QPointF(posn[2], posn[3]))
        painter = phelper.painter(self, posn, clip=clip)
        with painter:
            # adjustable positions for the lines
            arrowsize = s.get('arrowSize').convert(painter)

            # drawing settings for line
            if not s.Line.hide:
                painter.setPen( s.get('Line').makeQPen(painter) )
            else:
                painter.setPen( qt.QPen(qt.Qt.NoPen) )

            # settings for fill
            if not s.Fill.hide:
                painter.setBrush( s.get('Fill').makeQBrush(painter) )
            else:
                painter.setBrush( qt.QBrush() )

            # iterate over positions
            scaling = posn[2]-posn[0]
            if s.mode == 'length-angle':
                lines = self._computeLinesLengthAngle(posn, scaling)
            else:
                lines = self._computeLinesPointToPoint(posn)

            if lines is None:
                return

            controlgraphitems = []
            for index, (x, y, l, a) in enumerate(lines):

                utils.plotLineArrow(
                    painter, x, y, l, a,
                    arrowsize=arrowsize,
                    arrowleft=s.arrowleft,
                    arrowright=s.arrowright)

                if isnotdataset:
                    cgi = controlgraph.ControlLine(
                        self, phelper, x, y,
                        x + l*math.cos(a/180.*math.pi),
                        y + l*math.sin(a/180.*math.pi))
                    cgi.index = index
                    cgi.widgetposn = posn
                    controlgraphitems.append(cgi)

        phelper.setControlGraph(self, controlgraphitems)

    def updateControlItem(self, cgi, pt1, pt2):
        """If control items are moved, update line."""
        s = self.settings

        # calculate new position coordinate for item
        xpos, ypos = self._getGraphCoords(
            cgi.widgetposn, [pt1[0], pt1[0]+1], [pt1[1], pt1[1]+1])
        if xpos is None or ypos is None:
            return

        x = s.get('xPos').getFloatArray(self.document)
        y = s.get('yPos').getFloatArray(self.document)
        if x is None or y is None:
            return
        x, y = list(x), list(y)

        idx = min(cgi.index, len(x)-1)
        if not N.allclose(x[idx], xpos[0]):
            x[idx] = utils.round2delt(xpos[0], xpos[1])
        idx = min(cgi.index, len(y)-1)
        if not N.allclose(y[idx], ypos[0]):
            y[idx] = utils.round2delt(ypos[0], ypos[1])
        operations = [
            document.OperationSettingSet(s.get('xPos'), x),
            document.OperationSettingSet(s.get('yPos'), y),
        ]

        if s.mode == 'length-angle':
            # convert 2nd point to length, angle
            def la(ptx, pty):
                length = (
                    math.sqrt( (ptx-pt1[0])**2 + (pty-pt1[1])**2 ) /
                    (cgi.widgetposn[2]-cgi.widgetposn[0]) )
                angle = (
                    (math.atan2( pty-pt1[1], ptx-pt1[0] ) *
                     180. / math.pi) % 360. )
                return length, angle
            length, angle = la(pt2[0], pt2[1])
            # calculate length angle for neighbouring point, to get delta
            ldelt, adelt = la(pt2[0]+1, pt2[1]+1)

            # update values
            l = s.get('length').getFloatArray(self.document)
            a = s.get('angle').getFloatArray(self.document)
            if l is None or a is None:
                return
            l, a = list(l), list(a)

            idx = min(cgi.index, len(l)-1)
            if abs(l[idx]-length) > 1e-8:
                l[idx] = utils.round2delt(length, ldelt)
            idx = min(cgi.index, len(a)-1)
            if abs(a[idx]-angle) > 1e-8:
                a[idx] = utils.round2delt(angle, adelt)

            operations += [
                document.OperationSettingSet(s.get('length'), l),
                document.OperationSettingSet(s.get('angle'), a),
            ]
        else:
            xpos2, ypos2 = self._getGraphCoords(
                cgi.widgetposn, [pt2[0], pt2[0]+1], [pt2[1], pt2[1]+1])
            if xpos2 is not None and ypos2 is not None:
                x2 = s.get('xPos2').getFloatArray(self.document)
                y2 = s.get('yPos2').getFloatArray(self.document)
                if x2 is None or y2 is None:
                    return
                x2, y2 = list(x2), list(y2)

                idx = min(cgi.index, len(x2)-1)
                if not N.allclose(x2[idx], xpos2[0]):
                    x2[idx] = utils.round2delt(xpos2[0], xpos2[1])
                idx = min(cgi.index, len(y2)-1)
                if not N.allclose(y2[idx], ypos2[0]):
                    y2[idx] = utils.round2delt(ypos2[0], ypos2[1])

                operations += [
                    document.OperationSettingSet(s.get('xPos2'), x2),
                    document.OperationSettingSet(s.get('yPos2'), y2)
                ]

        self.document.applyOperation(
            document.OperationMultiple(operations, descr=_('adjust lines')) )

document.thefactory.register(Line)



#    Copyright (C) 2013 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import collections
import re

import numpy as N
from .. import qtall as qt
from .. import document
from .. import datasets
from .. import utils
from . import base
from . import fits_hdf5_helpers

def _(text, disambiguation=None, context="Import_HDF5"):
    return qt.QCoreApplication.translate(context, text, disambiguation)

h5py = None
def inith5py():
    global h5py
    try:
        import h5py
    except ImportError:
        raise RuntimeError(
            "Cannot load Python h5py module. "
            "Please install before loading documents using HDF5 data.")

def bconv(s):
    """Hack for h5py byte problem with python3.
    https://github.com/h5py/h5py/issues/379

    Byte string attributes are not converted to normal strings."""

    if isinstance(s, bytes):
        return s.decode('utf-8', 'replace')
    return s

def auto_deref_attr(attr, attrs, grp):
    """Automatic dereference any attributes which are references."""
    val = attrs[attr]
    # convert reference to a dataset
    if isinstance(val, h5py.Reference):
        # have to find root to dereference reference
        root = grp
        while root.name != '/':
            root = root.parent
        val = root[val]
    # convert dataset to an array
    if isinstance(val, h5py.Dataset):
        val = N.array(val)
    return bconv(val)

class ImportParamsHDF5(base.ImportParamsBase):
    """HDF5 file import parameters.

    Additional parameters:
     items: list of datasets and items to import
     namemap: map hdf datasets to veusz names
     slices: dict to map hdf names to slices
     twodranges: map hdf names to 2d range (minx, miny, maxx, maxy)
     twod_as_oned: set of hdf names to read 2d dataset as 1d dataset
     convert_datetime: map float or strings to datetime
    """

    defaults = {
        'items': None,
        'namemap': None,
        'slices': None,
        'twodranges': None,
        'twod_as_oned': None,
        'convert_datetime': None,
    }
    defaults.update(base.ImportParamsBase.defaults)

class LinkedFileHDF5(base.LinkedFileBase):
    """Links a HDF5 file to the data."""

    def createOperation(self):
        """Return operation to recreate self."""
        return OperationDataImportHDF5

    def saveToFile(self, fileobj, relpath=None):
        """Save the link to the document file."""
        self._saveHelper(
            fileobj,
            'ImportFileHDF5',
            ('filename', 'items'),
            relpath=relpath)

class _DataRead:
    """Data read from file during import.

    This is so we can store the original name and options stored in
    attributes from the file.
    """
    def __init__(self, origname, data, options):
        self.origname = origname
        self.data = data
        self.options = options

class OperationDataImportHDF5(base.OperationDataImportBase):
    """Import 1d, 2d, text or nd data from a HDF5 file."""

    descr = _("import HDF5 file")

    def readDataset(self, dataset, dsattrs, dsname, dsread):
        """Given hdf5 dataset, its attributes and name, get data and
        set it in dict dsread.

        dsread maps names to _DataRead object
        """

        # store options associated with dataset
        options = {}
        for a in dsattrs:
            if a[:4] == "vsz_":
                options[a] = auto_deref_attr(a, dsattrs, dataset)

        # find name for dataset
        if (self.params.namemap is not None and
            dsname in self.params.namemap ):
            name = self.params.namemap[dsname]
        else:
            if "vsz_name" in options:
                # override name using attribute
                name = options["vsz_name"]
            else:
                name = dsname.split("/")[-1].strip()

        # use full path if dataset already exists
        if name in dsread:
            name = dsname.strip()

        try:
            # implement slicing
            aslice = None
            if "vsz_slice" in options:
                s = fits_hdf5_helpers.convertTextToSlice(
                    options["vsz_slice"], len(dataset.shape))
                if s != -1:
                    aslice = s
            if self.params.slices and dsname in self.params.slices:
                aslice = self.params.slices[dsname]

            # finally return data
            objdata = fits_hdf5_helpers.convertDatasetToObject(
                dataset, aslice)
            dsread[name] = _DataRead(dsname, objdata, options)

        except fits_hdf5_helpers.ConvertError:
            pass

    def walkFile(self, item, dsread, names=None):
        """Walk an hdf file, adding datasets to dsread.

        If names is set to a list, only read names from list given
        """

        if isinstance(item, h5py.Dataset):
            try:
                dtype = item.dtype
            except TypeError:
                # not supported by h5py
                return

            if dtype.kind == 'V':
                # compound dataset - walk columns
                if not names:
                    names = item.dtype.names

                for name in names:
                    attrs = fits_hdf5_helpers.filterAttrsByName(item.attrs, name)
                    self.readDataset(
                        item[name], attrs, item.name+"/"+name, dsread)
            else:
                self.readDataset(item, item.attrs, item.name, dsread)

        elif isinstance(item, h5py.Group):
            if not names:
                names = sorted(item.keys())

            for dsname in names:
                try:
                    child = item[dsname]
                except KeyError:
                    # this does happen!
                    continue
                self.walkFile(child, dsread)

    def readDataFromFile(self):
        """Read data from hdf5 file and return a dict of names to data."""

        dsread = {}
        with h5py.File(self.params.filename, "r") as hdff:
            for hi in self.params.items:
                # lookup group/dataset in file
                names = [x for x in hi.split("/") if x != ""]
                node = hdff

                # Repeat until we get a dataset. Note: if we get a
                # dataset which names is not empty, this is a table
                # column, so we pass the remainder of names to
                # walkFile
                while names and not isinstance(node, h5py.Dataset):
                    node = node[names[0]]
                    names.pop(0)

                self.walkFile(node, dsread, names=names)
        return dsread

    def collectErrorBarDatasets(self, dsread):
        """Identify error bar datasets and separate out.
        Returns error bar datasets."""

        # separate out datasets with error bars
        # this a defaultdict of defaultdict with None as default
        errordatasets = collections.defaultdict(
            lambda: collections.defaultdict(lambda: None))
        for name in list(dsread):
            dr = dsread[name]
            ds = dr.data
            if not isinstance(ds, N.ndarray) or len(ds.shape) != 1:
                # skip non-numeric or 2d datasets
                continue

            for err in ('+', '-', '+-'):
                ln = len(err)+3
                if name[-ln:] == (' (%s)' % err):
                    refname = name[:-ln].strip()
                    if refname in dsread:
                        errordatasets[refname][err] = ds
                        del dsread[name]
                        break

        return errordatasets

    def numericDataToDataset(self, name, dread, errordatasets):
        """Convert numeric data to a veusz dataset."""

        data = dread.data

        ds = None
        if data.ndim == 1:
            if ( (self.params.convert_datetime and
                  dread.origname in self.params.convert_datetime) or
                 "vsz_convert_datetime" in dread.options ):

                try:
                    mode = self.params.convert_datetime[dread.origname]
                except (TypeError, KeyError):
                    mode = dread.options["vsz_convert_datetime"]

                if mode == 'unix':
                    data = utils.floatUnixToVeusz(data)
                ds = datasets.DatasetDateTime(data)

            else:
                # Standard 1D Import
                # handle any possible error bars
                args = {
                    'data': data,
                    'serr': errordatasets[name]['+-'],
                    'nerr': errordatasets[name]['-'],
                    'perr': errordatasets[name]['+']
                }

                # find minimum length and cut down if necessary
                minlen = min(
                    [len(d) for d in args.values() if d is not None])
                for a in list(args):
                    if args[a] is not None and len(args[a]) > minlen:
                        args[a] = args[a][:minlen]

                ds = datasets.Dataset(**args)

        elif data.ndim == 2:
            # 2D dataset
            if ( ((self.params.twod_as_oned and
                   dread.origname in self.params.twod_as_oned) or
                  dread.options.get("vsz_twod_as_oned") ) and
                 data.shape[1] in (2,3) ):
                # actually a 1D dataset in disguise
                if data.shape[1] == 2:
                    ds = datasets.Dataset(data=data[:,0], serr=data[:,1])
                else:
                    ds = datasets.Dataset(
                        data=data[:,0], perr=data[:,1], nerr=data[:,2])
            else:
                # this really is a 2D dataset

                attrs = {}
                # find any ranges
                if "vsz_range" in dread.options:
                    r = dread.options["vsz_range"]
                    attrs["xrange"] = (r[0], r[2])
                    attrs["yrange"] = (r[1], r[3])
                for attr in (
                        "xrange", "yrange", "xcent", "ycent",
                        "xedge", "yedge"):
                    if "vsz_"+attr in dread.options:
                        attrs[attr] = dread.options.get("vsz_"+attr)

                if ( self.params.twodranges and
                     dread.origname in self.params.twodranges ):
                    r = self.params.twodranges[dread.origname]
                    attrs["xrange"] = (r[0], r[2])
                    attrs["yrange"] = (r[1], r[3])

                # create the object
                ds = datasets.Dataset2D(data, **attrs)

        else:
            # N-dimensional dataset
            ds = datasets.DatasetND(data)

        return ds

    def textDataToDataset(self, name, dread):
        """Convert textual data to a veusz dataset."""

        data = dread.data

        if ( (self.params.convert_datetime and
              dread.origname in self.params.convert_datetime) or
             "vsz_convert_datetime" in dread.options ):

            try:
                fmt = self.params.convert_datetime[dread.origname]
            except (TypeError, KeyError):
                fmt = dread.options["vsz_convert_datetime"]

            if fmt.strip() == 'iso':
                fmt = 'YYYY-MM-DD|T|hh:mm:ss'

            try:
                datere = re.compile(utils.dateStrToRegularExpression(fmt))
            except Exception:
                raise base.ImportingError(
                    _("Could not interpret date-time syntax '%s'") % fmt)

            dout = N.empty(len(data), dtype=N.float64)
            for i, ditem in enumerate(data):
                ditem = bconv(ditem)
                try:
                    match = datere.match(ditem)
                    val = utils.dateREMatchToDate(match)
                except ValueError:
                    val = N.nan
                dout[i] = val

            ds = datasets.DatasetDateTime(dout)

        else:
            # unfortunately byte strings are returned in py3
            tdata = [bconv(d) for d in dread.data]

            # standard text dataset
            ds = datasets.DatasetText(tdata)

        return ds

    def doImport(self):
        """Do the import."""

        inith5py()
        par = self.params

        dsread = self.readDataFromFile()

        # find datasets which are error bars
        errordatasets = self.collectErrorBarDatasets(dsread)

        if par.linked:
            linkedfile = LinkedFileHDF5(par)
        else:
            linkedfile = None

        # create the veusz output datasets
        for name, dread in dsread.items():
            if isinstance(dread.data, N.ndarray):
                # numeric
                ds = self.numericDataToDataset(name, dread, errordatasets)
            else:
                # text
                ds = self.textDataToDataset(name, dread)

            if ds is None:
                # error above
                continue

            ds.linked = linkedfile

            # finally set dataset in document
            fullname = par.prefix + name + par.suffix
            self.outdatasets[fullname] = ds

def ImportFileHDF5(comm, filename,
                   items,
                   namemap=None,
                   slices=None,
                   twodranges=None,
                   twod_as_oned=None,
                   convert_datetime=None,
                   prefix='', suffix='',
                   renames=None,
                   linked=False):
    """Import data from a HDF5 file

    items is a list of groups and datasets which can be imported.
    If a group is imported, all child datasets are imported.

    namemap maps an input dataset to a veusz dataset name. Special
    suffixes can be used on the veusz dataset name to indicate that
    the dataset should be imported specially.

    'foo (+)': import as +ve error for dataset foo
    'foo (-)': import as -ve error for dataset foo
    'foo (+-)': import as symmetric error for dataset foo

    slices is an optional dict specifying slices to be selected when
    importing. For each dataset to be sliced, provide a tuple of
    values, one for each dimension. The values should be a single
    integer to select that index, or a tuple (start, stop, step),
    where the entries are integers or None.

    twodranges is an optional dict giving data ranges for 2d
    datasets. It maps names to (minx, miny, maxx, maxy).

    twod_as_oned: optional set containing 2d datasets to attempt to
    read as 1d

    convert_datetime should be a dict mapping hdf name to specify
    date/time importing
      for a 1d numeric dataset
        if this is set to 'veusz', this is the number of seconds since
          2009-01-01
        if this is set to 'unix', this is the number of seconds since
          1970-01-01
       for a text dataset, this should give the format of the date/time,
          e.g. 'YYYY-MM-DD|T|hh:mm:ss' or 'iso' for iso format

    renames is a dict mapping old to new dataset names, to be renamed
    after importing

    linked specifies that the dataset is linked to the file.

    Attributes can be used in datasets to override defaults:
     'vsz_name': set to override name for dataset in veusz
     'vsz_slice': slice on importing (use format "start:stop:step,...")
     'vsz_range': should be 4 item array to specify x and y ranges:
                  [minx, miny, maxx, maxy]
     'vsz_xrange' / 'vsz_yrange': individual ranges for x and y
     'vsz_xcent' / 'vsz_ycent': arrays giving the centres of pixels
     'vsz_xedge' / 'vsz_yedge': arrays giving the edges of pixels
     'vsz_twod_as_oned': treat 2d dataset as 1d dataset with errors
     'vsz_convert_datetime': treat as date/time, set to one of the values
                             above.
    References to other datasets can be provided in thes attributes.

    For compound datasets these attributes can be given on a
    per-column basis using attribute names
    vsz_attributename_columnname.

    Returns: list of imported datasets
    """

    # lookup filename
    realfilename = comm.findFileOnImportPath(filename)

    params = ImportParamsHDF5(
        filename=realfilename,
        items=items,
        namemap=namemap,
        slices=slices,
        twodranges=twodranges,
        twod_as_oned=twod_as_oned,
        convert_datetime=convert_datetime,
        prefix=prefix, suffix=suffix,
        renames=renames,
        linked=linked)
    op = OperationDataImportHDF5(params)
    comm.document.applyOperation(op)

    if comm.verbose:
        print("Imported datasets %s" % ', '.join(op.outnames))
    return op.outnames

document.registerImportCommand("ImportFileHDF5", ImportFileHDF5)



# data editting dialog

#    Copyright (C) 2005 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Module for implementing dialog box for viewing/editing data."""

import numpy as N

from .. import qtall as qt
from .. import document
from .. import datasets
from .. import setting
from ..qtwidgets.datasetbrowser import DatasetBrowser
from .veuszdialog import VeuszDialog, recreate_register

def _(text, disambiguation=None, context="DataEditDialog"):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class DatasetTableModel1D(qt.QAbstractTableModel):
    """Provides access to editing and viewing of datasets."""

    def __init__(self, parent, document, datasetname):
        qt.QAbstractTableModel.__init__(self, parent)

        self.document = document
        self.dsname = datasetname

        document.signalModified.connect(self.slotDocumentModified)

    def rowCount(self, parent):
        """Return number of rows."""
        if parent.isValid():
            # docs say we should return zero
            return 0

        try:
            return len(self.document.data[self.dsname].data)+1
        except (KeyError, AttributeError):
            return 0

    def slotDocumentModified(self):
        """Called when document modified."""
        self.layoutChanged.emit()

    def columnCount(self, parent):
        """Return number of columns."""

        if parent.isValid():
            return 0
        try:
            ds = self.document.data[self.dsname]
        except KeyError:
            return 0
        return len( ds.column_descriptions )

    def data(self, index, role):
        """Return data for index."""
        # get dataset
        ds = self.document.data[self.dsname]
        if ds is not None:
            # select correct part of dataset
            data = getattr(ds, ds.columns[index.column()])
        if ds is not None and data is not None and role in (
            qt.Qt.DisplayRole, qt.Qt.EditRole):
            # blank row at end of data
            if index.row() == len(data):
                return None

            # convert data to data
            d = data[index.row()]
            return ds.uiDataItemToData(d)

        # empty entry
        return None

    def headerData(self, section, orientation, role):
        """Return row numbers or column names."""

        try:
            ds = self.document.data[self.dsname]
        except KeyError:
            return None

        if role == qt.Qt.DisplayRole:
            if orientation == qt.Qt.Horizontal:
                # column names
                return ds.column_descriptions[section]
            else:
                if section == len(ds.data):
                    return "+"
                # return row numbers
                return section+1

        return None

    def flags(self, index):
        """Update flags to say that items are editable."""
        if index.isValid():
            f = qt.QAbstractTableModel.flags(self, index)
            ds = self.document.data.get(self.dsname)
            if ds is not None and ds.editable:
                f |= qt.Qt.ItemIsEditable
            return f
        return qt.Qt.ItemIsEnabled

    def removeRows(self, row, count):
        """Remove rows."""
        self.document.applyOperation(
            document.OperationDatasetDeleteRow(self.dsname, row, count))

    def insertRows(self, row, count):
        """Remove rows."""
        self.document.applyOperation(
            document.OperationDatasetInsertRow(self.dsname, row, count))

    def setData(self, index, value, role):
        """Called to set the data."""

        if not index.isValid() or role != qt.Qt.EditRole:
            return False

        row = index.row()
        column = index.column()
        ds = self.document.data[self.dsname]
        data = getattr(ds, ds.columns[index.column()])

        # add new column if necessary
        ops = document.OperationMultiple([], descr=_('set value'))
        if data is None:
            ops.addOperation(
                document.OperationDatasetAddColumn(
                    self.dsname, ds.columns[column]))

        # add a row if necessary
        if row == len(ds.data):
            ops.addOperation(
                document.OperationDatasetInsertRow(self.dsname, row, 1))

        # update if conversion okay
        try:
            val = ds.uiConvertToDataItem(value)
        except ValueError:
            return False

        ops.addOperation(
            document.OperationDatasetSetVal(
                self.dsname, ds.columns[column], row, val))
        try:
            self.document.applyOperation(ops)
        except RuntimeError:
            return False
        return True

class DatasetTableModelMulti(qt.QAbstractTableModel):
    """Edit multiple datasets simultaneously with a spreadsheet-like style."""

    def __init__(self, parent, document, datasetnames):
        qt.QAbstractTableModel.__init__(self, parent)

        self.document = document
        self.dsnames = datasetnames
        document.signalModified.connect(self.slotDocumentModified)

        self.changeset = -1
        self.rows = 0

    def updateCounts(self):
        """Count rows and columns."""

        self.changeset = self.document.changeset

        rows = 0
        rowcounts = self.rowcounts = []
        colcounts = self.colcounts = []
        colattrs = self.colattrs = []

        for dsidx, name in enumerate(self.dsnames):
            if name not in self.document.data:
                continue
            dataset = self.document.data[name]
            if (not hasattr(dataset, 'data') or
                not hasattr(dataset, 'columns') or
                dataset.dimensions != 1):
                continue

            r = len(dataset.data)+1
            rowcounts.append(r)
            rows = max(rows, r)

            attr = []
            for colidx, col in enumerate(dataset.columns):
                data = getattr(dataset, col)
                if data is not None:
                    attr.append( (name, col, dsidx, colidx) )
            colcounts.append( len(attr) )
            colattrs += attr

        self.rows = rows

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        if self.changeset != self.document.changeset:
            self.updateCounts()
        return self.rows

    def columnCount(self, parent):
        if parent.isValid():
            return 0
        if self.changeset != self.document.changeset:
            self.updateCounts()
        return len(self.colattrs)

    def slotDocumentModified(self):
        self.updateCounts()
        self.layoutChanged.emit()

    def data(self, index, role):
        """Return data for index."""

        dsname, colname, dsidx, colidx = self.colattrs[index.column()]
        ds = self.document.data[dsname]
        data = getattr(ds, colname)

        if role == qt.Qt.DisplayRole:
            if index.row() < self.rowcounts[dsidx]-1:
                # convert data to Data
                d = data[index.row()]
                return ds.uiDataItemToData(d)

        # empty entry
        return None

    def headerData(self, section, orientation, role):
        """Return row numbers or column names."""

        if role == qt.Qt.DisplayRole:
            if orientation == qt.Qt.Horizontal:
                # column names
                dsname, colname, dsidx, colidx = self.colattrs[section]
                ds = self.document.data[dsname]
                descr = ds.column_descriptions[colidx]
                header = dsname + '\n' + descr
                return header
            else:
                # return row numbers
                if section == self.rows-1:
                    return "+"
                return section+1

        return None

    def flags(self, index):
        """Update flags to say that items are editable."""
        if index.isValid():
            f = qt.QAbstractTableModel.flags(self, index)
            dsname, colname, dsidx, colidx = self.colattrs[index.column()]
            ds = self.document.data.get(dsname)
            if ds is not None and ds.editable:
                f |= qt.Qt.ItemIsEditable
            return f
        return qt.Qt.ItemIsEnabled

    def setData(self, index, value, role):
        """Validate and set data in dataset."""

        if not index.isValid() or role != qt.Qt.EditRole:
            return False

        row = index.row()
        column = index.column()
        dsname, colname, dsidx, colidx = self.colattrs[column]
        ds = self.document.data[dsname]

        ops = document.OperationMultiple([], descr=_('set value'))
        if row >= self.rowcounts[dsidx]-1:
            # add number of rows required to add new value below
            ops.addOperation(
                document.OperationDatasetInsertRow(
                    dsname, self.rowcounts[dsidx]-1,
                    row+1-self.rowcounts[dsidx]+1))

        # convert text to value
        try:
            val = ds.uiConvertToDataItem(value)
        except ValueError:
            return False

        ops.addOperation(
            document.OperationDatasetSetVal(dsname, colname, row, val))

        try:
            self.document.applyOperation(ops)
            return True
        except RuntimeError:
            return False

    def insertRows(self, row, count):
        ops = []
        for i, name in enumerate(self.dsnames):
            if self.rowcounts[i]-1 >= row:
                ops.append(
                    document.OperationDatasetInsertRow(name, row, count))
        self.document.applyOperation(
            document.OperationMultiple(ops, _('insert row(s)')))

    def removeRows(self, row, count):
        ops = []
        for i, name in enumerate(self.dsnames):
            if self.rowcounts[i]-1 >= row:
                ops.append(
                    document.OperationDatasetDeleteRow(name, row, count))
        self.document.applyOperation(
            document.OperationMultiple(ops, _('delete row(s)')))

class DatasetTableModel2D(qt.QAbstractTableModel):
    """A 2D dataset model."""

    def __init__(self, parent, document, datasetname):
        qt.QAbstractTableModel.__init__(self, parent)

        self.document = document
        self.dsname = datasetname
        self.updatePixelCoords()
        document.signalModified.connect(self.slotDocumentModified)

    def updatePixelCoords(self):
        """Get coordinates at edge of grid."""
        self.xedge = self.yedge = self.xcent = self.ycent = []
        ds = self.document.data.get(self.dsname)
        if ds and ds.dimensions==2:
            self.xcent, self.ycent = ds.getPixelCentres()
            self.xedge, self.yedge = ds.getPixelEdges()

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        try:
            data = self.document.data[self.dsname].data
        except KeyError:
            return 0
        if data is not None and data.ndim==2:
            return data.shape[0]
        else:
            return 0

    def columnCount(self, parent):
        if parent.isValid():
            return 0
        try:
            data = self.document.data[self.dsname].data
        except KeyError:
            return 0
        if data is not None and data.ndim==2:
            return data.shape[1]
        else:
            return 0

    def data(self, index, role):
        if role == qt.Qt.DisplayRole:
            # get data (note y is reversed, sigh)
            try:
                data = self.document.data[self.dsname].data
            except KeyError:
                return None
            if data is not None and data.ndim==2:
                try:
                    num = data[data.shape[0]-index.row()-1, index.column()]
                    return float(num)
                except IndexError:
                    pass

        return None

    def headerData(self, section, orientation, role):
        """Return headers at top."""

        ds = self.document.data.get(self.dsname)
        if ds.dimensions != 2:
            return None

        xaxis = orientation == qt.Qt.Horizontal

        # note: y coordinates are upside down (high y is at top)
        if ds is not None and role == qt.Qt.DisplayRole:
            v = self.xcent[section] if xaxis else self.ycent[
                len(self.ycent)-section-1]
            return '%i (%s)' % (
                len(self.ycent)-section, setting.ui_floattostring(v, maxdp=4))

        elif ds is not None and role == qt.Qt.ToolTipRole:
            v1 = self.xedge[section] if xaxis else self.yedge[
                len(self.yedge)-section-2]
            v2 = self.xedge[section+1] if xaxis else self.yedge[
                len(self.yedge)-section-1]
            return '%s\u2013%s' % (
                setting.ui_floattostring(v1), setting.ui_floattostring(v2))

        return None

    def flags(self, index):
        """Update flags to say that items are editable."""
        if not index.isValid():
            return qt.Qt.ItemIsEnabled
        else:
            f = qt.QAbstractTableModel.flags(self, index)
            ds = self.document.data.get(self.dsname)
            if ds is not None and ds.editable:
                f |= qt.Qt.ItemIsEditable
            return f

    def slotDocumentModified(self):
        """Called when document modified."""
        self.updatePixelCoords()
        self.layoutChanged.emit()

    def setData(self, index, value, role):
        """Called to set the data."""

        if not index.isValid() or role != qt.Qt.EditRole:
            return False

        ds = self.document.data[self.dsname]
        row = ds.data.shape[0]-index.row()-1
        col = index.column()

        # update if conversion okay
        try:
            val = ds.uiConvertToDataItem(value)
        except ValueError:
            return False

        op = document.OperationDatasetSetVal2D(
            self.dsname, row, col, val)
        self.document.applyOperation(op)
        return True

class DatasetTableModelND(qt.QAbstractTableModel):
    """An ND dataset model."""

    def __init__(self, parent, document, datasetname):
        qt.QAbstractTableModel.__init__(self, parent)

        self.document = document
        self.dsname = datasetname
        document.signalModified.connect(self.slotDocumentModified)

    def rowCount(self, parent):
        if parent.isValid():
            return 0
        try:
            data = self.document.data[self.dsname].data
        except KeyError:
            return 0
        return 0 if data is None else data.size

    def columnCount(self, parent):
        if parent.isValid():
            return 0
        try:
            data = self.document.data[self.dsname].data
        except KeyError:
            return 0
        return 1 if data is not None else 0

    def data(self, index, role):
        """Items in array."""
        if role == qt.Qt.DisplayRole:
            try:
                data = self.document.data[self.dsname].data
            except KeyError:
                return None
            if data is not None:
                try:
                    num = N.ravel(data)[index.row()]
                    return float(num)
                except IndexError:
                    pass
        return None

    def headerData(self, section, orientation, role):
        """Return headers at top."""

        ds = self.document.data.get(self.dsname)
        if ds is None:
            return None

        if ds is not None and role == qt.Qt.DisplayRole:
            if orientation == qt.Qt.Horizontal:
                return _('Value')
            else:
                idx = N.unravel_index(section, ds.data.shape)
                txt = ','.join( [str(v+1) for v in idx] )
                return txt
        return None

    def slotDocumentModified(self):
        """Called when document modified."""
        self.layoutChanged.emit()

class ViewDelegate(qt.QStyledItemDelegate):
    """Delegate for fixing double editing.
    Normal editing uses double spin box, which is inappropriate
    """

    def createEditor(self, parent, option, index):
        if type(index.data()) is float:
            return qt.QLineEdit(parent)
        else:
            return qt.QStyledItemDelegate.createEditor(
                self, parent, option, index)

    def setEditorData(self, editor, index):
        """Override setData to use correct formatting."""
        if type(index.data()) is float:
            txt = setting.ui_floattostring(index.data())
            editor.setText(txt)
        else:
            qt.QStyledItemDelegate.setEditorData(self, editor, index)

class DataEditDialog(VeuszDialog):
    """Dialog for editing and rearranging data sets."""

    def __init__(self, parent, document):
        VeuszDialog.__init__(self, parent, 'dataedit.ui')
        self.document = document

        # set up dataset list
        self.dsbrowser = DatasetBrowser(document, parent, parent)
        self.dsbrowser.setToolTip(
            _('Select multiple datasets to edit simultaneously'))
        self.splitter.insertWidget(0, self.dsbrowser)

        self.deligate = ViewDelegate()
        self.datatableview.setItemDelegate(self.deligate)

        # actions for data table
        for text, slot in (
            (_('Copy'), self.slotCopy),
            (_('Delete row'), self.slotDeleteRow),
            (_('Insert row'), self.slotInsertRow),
        ):
            act = qt.QAction(text, self)
            act.triggered.connect(slot)
            self.datatableview.addAction(act)
        self.datatableview.setContextMenuPolicy( qt.Qt.ActionsContextMenu )

        # layout edit dialog improvement
        self.splitter.setStretchFactor(0, 3)
        self.splitter.setStretchFactor(1, 4)

        # don't want text to look editable or special
        self.linkedlabel.setFrameShape(qt.QFrame.NoFrame)
        self.linkedlabel.viewport().setBackgroundRole(qt.QPalette.Window)

        # document changes
        document.signalModified.connect(self.slotDocumentModified)

        # select first item, if any or initialise if none
        if len(self.document.data) > 0:
            self.selectDataset( sorted(self.document.data)[0] )
        else:
            self.slotDatasetsSelected([])

        self.dsbrowser.navtree.selecteddatasets.connect(
            self.slotDatasetsSelected)

        # connect buttons
        for btn, slot in (
                (self.deletebutton, self.slotDatasetDelete),
                (self.unlinkbutton, self.slotDatasetUnlink),
                (self.duplicatebutton, self.slotDatasetDuplicate),
                (self.importbutton, self.slotDatasetImport),
                (self.createbutton, self.slotDatasetCreate),
                (self.editbutton, self.slotDatasetEdit),
        ):
            btn.clicked.connect(slot)

        # menu for new button
        self.newmenu = qt.QMenu()
        for text, slot in (
                (_('Numerical dataset'), self.slotNewNumericalDataset),
                (_('Text dataset'), self.slotNewTextDataset),
                (_('Date/time dataset'), self.slotNewDateDataset)
        ):
            a = self.newmenu.addAction(text)
            a.triggered.connect(slot)
        self.newbutton.setMenu(self.newmenu)

    def slotDatasetsSelected(self, names):
        """Called when a new dataset is selected."""

        # FIXME: Make readonly models readonly!!
        model = None
        if len(names) == 1:
            # get selected dataset
            ds = self.document.data[names[0]]

            # make model for dataset
            if ds.dimensions == 1:
                model = DatasetTableModel1D(self, self.document, names[0])
            elif ds.dimensions == 2:
                model = DatasetTableModel2D(self, self.document, names[0])
            elif ds.dimensions == -1:
                model = DatasetTableModelND(self, self.document, names[0])
        elif len(names) > 1:
            model = DatasetTableModelMulti(self, self.document, names)

        # disable context menu if no menu
        for a in self.datatableview.actions():
            a.setEnabled(model is not None)

        self.datatableview.setModel(model)
        self.setUnlinkState()

    def setUnlinkState(self):
        """Enable the unlink button correctly."""

        linkinfo = []
        canunlink = []
        canedit = []
        names = self.dsbrowser.navtree.getSelectedDatasets()
        for name in names:
            ds = self.document.data[name]
            canunlink.append(ds.canUnlink())
            if len(names) > 1:
                linkinfo.append(name)
            linkinfo.append(ds.linkedInformation())
            canedit.append(type(ds) in recreate_register)

        self.editbutton.setVisible(any(canedit))
        self.unlinkbutton.setEnabled(any(canunlink))
        self.linkedlabel.setText('\n'.join(linkinfo))
        self.deletebutton.setEnabled(bool(names))
        self.duplicatebutton.setEnabled(bool(names))

    def slotDocumentModified(self):
        """Set unlink status when document modified."""
        self.setUnlinkState()

    def selectDataset(self, dsname):
        """Select dataset with name given."""
        self.dsbrowser.navtree.selectDataset(dsname)
        self.slotDatasetsSelected([dsname])

    def slotDatasetDelete(self):
        """Delete selected dataset."""
        dsnames = self.dsbrowser.navtree.getSelectedDatasets()
        self.document.applyOperation(
            document.OperationMultiple(
                [document.OperationDatasetDelete(n) for n in dsnames],
                descr=_('delete dataset(s)')))

    def slotDatasetUnlink(self):
        """Allow user to remove link to file or other datasets."""
        ops = []
        for name in self.dsbrowser.navtree.getSelectedDatasets():
            d = self.document.data[name]
            if d.linked is not None:
                ops.append(document.OperationDatasetUnlinkFile(name))
            elif d.canUnlink():
                ops.append(document.OperationDatasetUnlinkRelation(name))
        if ops:
            self.document.applyOperation(
                document.OperationMultiple(ops, _('unlink dataset(s)')))

    def slotDatasetDuplicate(self):
        """Duplicate selected datasets."""
        ops = []
        for name in self.dsbrowser.navtree.getSelectedDatasets():
            # generate new name for dataset
            newname = name + '_copy'
            index = 2
            while newname in self.document.data:
                newname = '%s_copy_%i' % (name, index)
                index += 1
            ops.append(
                document.OperationDatasetDuplicate(name, newname))
        if ops:
            self.document.applyOperation(
                document.OperationMultiple(ops, _('duplicate dataset(s)')))

    def slotDatasetImport(self):
        """Show import dialog."""
        self.mainwindow.slotDataImport()

    def slotDatasetCreate(self):
        """Show dataset creation dialog."""
        self.mainwindow.slotDataCreate()

    def slotDatasetEdit(self):
        """Reload dataset into dataset creation dialog."""
        for name in self.dsbrowser.navtree.getSelectedDatasets():
            dataset = self.document.data[name]
            try:
                recreate_register[type(dataset)](
                    self.mainwindow, self.document, dataset, name)
            except KeyError:
                pass

    def slotCopy(self):
        """Copy text from selection."""
        # get list of selected rows and columns
        selmodel = self.datatableview.selectionModel()
        model = self.datatableview.model()
        indices = []
        for index in selmodel.selectedIndexes():
            indices.append( (index.row(), index.column()) )
        indices.sort()

        # build up text stream for copying to clipboard
        lines = []
        rowitems = []
        lastrow = -1
        for row, column in indices:
            if row != lastrow:
                if rowitems:
                    # items are tab separated
                    lines.append( '\t'.join(rowitems) )
                    rowitems = []
                lastrow = row
            rowitems.append(
                str(model.createIndex(row, column).data()) )
        if rowitems:
            lines.append( '\t'.join(rowitems) )
        lines.append('')  # blank line at end
        lines = '\n'.join(lines)

        # put text on clipboard
        qt.QApplication.clipboard().setText(lines)

    def slotDeleteRow(self):
        """Delete the current row."""
        self.datatableview.model().removeRows(
            self.datatableview.currentIndex().row(), 1)

    def slotInsertRow(self):
        """Insert a new row."""
        self.datatableview.model().insertRows(
            self.datatableview.currentIndex().row(), 1)

    def slotNewNumericalDataset(self):
        """Add new value dataset."""
        self.newDataset( datasets.Dataset(data=[0.]) )

    def slotNewTextDataset(self):
        """Add new text dataset."""
        self.newDataset( datasets.DatasetText(data=['']) )

    def slotNewDateDataset(self):
        """Add new date dataset."""
        self.newDataset( datasets.DatasetDateTime(data=[]) )

    def newDataset(self, ds):
        """Add new dataset to document."""
        # get a name for dataset
        name = _('new dataset')
        if name in self.document.data:
            count = 1
            while name in self.document.data:
                name = _('new dataset %i') % count
                count += 1

        # add new dataset
        self.document.applyOperation(
            document.OperationDatasetSet(name, ds))

        self.dsbrowser.selectDataset(name)



#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Widgets are defined in this module."""

from .widget import Widget, Action
from .axis import Axis
from .axisbroken import AxisBroken
from .axisfunction import AxisFunction
from .graph import Graph
from .grid import Grid
from .plotters import GenericPlotter, FreePlotter
from .pickable import PickInfo
from .point import PointPlotter
from .function import FunctionPlotter
from .histo import Histo
from .textlabel import TextLabel
from .page import Page
from .root import Root
from .key import Key
from .fit import Fit
from .image import Image
from .contour import Contour
from .colorbar import ColorBar
from .shape import Shape, BoxShape, Rectangle, Ellipse, ImageFile, SVGFile
from .line import Line
from .bar import BarPlotter
from .polygon import Polygon
from .vectorfield import VectorField
from .boxplot import BoxPlot
from .polar import Polar
from .ternary import Ternary
from .nonorthpoint import NonOrthPoint
from .nonorthfunction import NonOrthFunction
from .scene3d import Scene3D
from .graph3d import Graph3D
from .axis3d import Axis3D
from .plotters3d import GenericPlotter3D
from .function3d import Function3D
from .point3d import Point3D
from .surface3d import Surface3D
from .covariance import Covariance
from .volume3d import Volume3D



#!/usr/bin/env python

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

'''Main Veusz executable.'''

import sys
import signal
import argparse
import re

import veusz
from veusz import qtall as qt
from veusz import utils

if sys.version_info[0] < 3:
    raise RuntimeError('Veusz only supports Python 3')

copyr='''Veusz %s

Copyright (C) Jeremy Sanders 2003-2021 <jeremy@jeremysanders.net>
 and contributors
Licenced under the GNU General Public Licence (version 2 or greater)
'''

splashcopyr='''<b><font color="purple">Veusz %s<br></font></b>
Copyright (C) Jeremy Sanders 2003-2021 and contributors<br>
Licenced under the GPL (version 2 or greater)
'''

def handleIntSignal(signum, frame):
    '''Ask windows to close if Ctrl+C pressed.'''
    qt.qApp.closeAllWindows()

def makeSplash(app):
    '''Make a splash screen logo.'''

    splash = qt.QSplashScreen()
    splash.setStyleSheet("background-color:white; color: black;")

    # draw logo on pixmap
    layout = qt.QVBoxLayout(splash)
    logo = qt.QLabel()
    logo.setPixmap(utils.getPixmap('logo.png'))
    logo.setAlignment(qt.Qt.AlignCenter)
    layout.addWidget(logo)

    # add copyright text
    message = qt.QLabel()
    message.setText(splashcopyr % utils.version())
    message.setAlignment(qt.Qt.AlignCenter)
    # increase size of font
    font = message.font()
    font.setPointSizeF(font.pointSize()*1.5)
    message.setFont(font)
    layout.addWidget(message)
    h = qt.QFontMetrics(font).height()
    layout.setContentsMargins(h,h,h,h)

    # Center the spash screen
    splash.setGeometry(5, 5, 100, 100)
    screen = qt.QDesktopWidget().screenGeometry()
    splash.move(
        (screen.width()-layout.sizeHint().width())//2,
        (screen.height()-layout.sizeHint().height())//2
    )

    # make sure dialog goes away - avoid problem if a message box pops
    # up before it is removed
    qt.QTimer.singleShot(2000, splash.hide)

    return splash

def excepthook(excepttype, exceptvalue, tracebackobj):
    '''Show exception dialog if an exception occurs.'''

    # exception dialog doesnt work if not in main thread, so we send
    # the exception to the application to display
    if qt.qApp.thread() is not qt.QThread.currentThread():
        qt.qApp.signalException.emit(excepttype, exceptvalue, tracebackobj)
        return

    sys.setrecursionlimit(sys.getrecursionlimit()+1000)

    from veusz.dialogs.exceptiondialog import ExceptionDialog
    if not isinstance(exceptvalue, utils.IgnoreException):
        # next exception is ignored to clear out the stack frame of the
        # previous exception - yuck
        d = ExceptionDialog((excepttype, exceptvalue, tracebackobj), None)
        d.exec_()

def listen(docs, quiet):
    '''For running with --listen option.'''
    from veusz.veusz_listen import openWindow
    openWindow(docs, quiet=quiet)

def export(exports, docs, options):
    '''A shortcut to load a set of files and export them.'''
    from veusz import document
    from veusz import utils

    # TODO: validate options
    opttxt = ', '.join(options) if options else ''

    for expfn, vsz in zip(exports, docs):
        doc = document.Document()
        ci = document.CommandInterpreter(doc)
        ci.Load(vsz)
        ci.run('Export(%s, %s)' % (repr(expfn), opttxt))

def convertArgsUnicode(args):
    '''Convert set of arguments to unicode (for Python 2).
    Arguments in argv use current file system encoding
    '''
    enc = sys.getfilesystemencoding()
    # bail out if not supported
    if enc is None:
        return args
    out = []
    for a in args:
        if isinstance(a, bytes):
            out.append(a.decode(enc))
        else:
            out.append(a)
    return out

class ImportThread(qt.QThread):
    '''Do import of main code within another thread.
    Main application runs when this is done
    '''
    def run(self):
        from veusz import setting
        from veusz import widgets
        from veusz import dataimport

class VeuszApp(qt.QApplication):
    """Event which can open mac files."""

    signalException = qt.pyqtSignal(object, object, object)

    def __init__(self):
        qt.QApplication.__init__(self, sys.argv)

        self.lastWindowClosed.connect(self.quit)
        self.signalException.connect(self.showException)

        # register a signal handler to catch ctrl+C
        signal.signal(signal.SIGINT, handleIntSignal)

        # parse command line options
        parser = argparse.ArgumentParser(
            description='Veusz scientific plotting package')
        parser.add_argument(
            '--version', action='version',
            version=copyr % utils.version())
        parser.add_argument(
            '--unsafe-mode',
            action='store_true',
            help='disable safety checks when running documents'
            ' or scripts')
        parser.add_argument(
            '--listen',
            action='store_true',
            help='read and execute Veusz commands from stdin,'
            ' replacing veusz_listen')
        parser.add_argument(
            '--quiet',
            action='store_true',
            help='if in listening mode, do not open a window but'
            ' execute commands quietly')
        parser.add_argument(
            '--export', action='append', metavar='FILE',
            help='export the next document to this'
            ' output image file, exiting when finished')
        parser.add_argument(
            '--export-option', action='append', metavar='VAL',
            help='add option when exporting file')
        parser.add_argument(
            '--embed-remote',
            action='store_true',
            help='(internal - not for external use)')
        parser.add_argument(
            '--plugin', action='append', metavar='FILE',
            help='load the plugin from the file given for '
            'the session')
        parser.add_argument(
            '--translation', metavar='FILE',
            help='load the translation .qm file given')
        parser.add_argument(
            'docs', metavar='FILE', nargs='*',
            help='document to load')

        self.args = args = parser.parse_args()

        args.docs = convertArgsUnicode(args.docs)

        # export files to make images
        if args.export:
            if len(args.export) != len(args.docs):
                parser.error(
                    'export option needs same number of documents and '
                    'output files')
            args.export = convertArgsUnicode(args.export)

        self.openeventfiles = []
        self.startupdone = False
        self.splash = None
        self.trans = None

    def openMainWindow(self, docs):
        """Open the main window with any loaded files."""
        from veusz.windows.mainwindow import MainWindow

        emptywins = []
        for w in self.topLevelWidgets():
            if isinstance(w, MainWindow) and w.document.isBlank():
                emptywins.append(w)

        if docs:
            # load in filenames given
            for filename in docs:
                if not emptywins:
                    MainWindow.CreateWindow(filename)
                elif filename:
                    emptywins[0].openFile(filename)
        else:
            # create blank window
            MainWindow.CreateWindow()

    def openPendingFiles(self):
        """If startup complete, open any files."""
        if self.startupdone:
            self.openMainWindow([None] + self.openeventfiles)
            del self.openeventfiles[:]
        else:
            qt.QTimer.singleShot(100, self.openPendingFiles)

    def event(self, event):
        """Handle events. This is the only way to get the FileOpen event.
        FileOpen is used by MacOS to open files.
        """
        if event.type() == qt.QEvent.FileOpen:
            self.openeventfiles.append(event.file())
            # need to wait until startup has finished
            qt.QTimer.singleShot(100, self.openPendingFiles)
            return True
        return qt.QApplication.event(self, event)

    def startup(self):
        """Do startup."""

        if not (self.args.listen or self.args.export):
            # show the splash screen on normal start
            self.splash = makeSplash(self)
            self.splash.show()

        self.thread = ImportThread()
        self.thread.finished.connect(self.slotStartApplication)
        self.thread.start()

    def slotStartApplication(self):
        """Start app, after modules imported."""

        args = self.args

        from veusz.utils import vzdbus, vzsamp
        vzdbus.setup()
        vzsamp.setup()

        # add text if we want to display an error after startup
        startuperrors = []

        from veusz import document
        from veusz import setting

        # install exception hook after thread has finished
        global defaultexcepthook
        defaultexcepthook = sys.excepthook
        sys.excepthook = excepthook

        # for people who want to run any old script
        setting.transient_settings['unsafe_mode'] = bool(
            args.unsafe_mode)

        # optionally load a translation
        txfile = args.translation or setting.settingdb['translation_file']
        if txfile:
            self.trans = qt.QTranslator()
            if self.trans.load(txfile):
                self.installTranslator(self.trans)
            else:
                startuperrors.append(
                    'Error loading translation "%s"' % txfile)

        # add directories to path
        if setting.settingdb['external_pythonpath']:
            # We want a list of items separated by colons
            # Unfortunately on windows there can be a colon and drive letter,
            # so we avoid splitting colons which look like a:\foo or B:/bar
            parts = re.findall(
                r'[A-Za-z]:[\\/][^:]+|[^:]+',
                setting.settingdb['external_pythonpath'])
            sys.path += list(parts)

        # load any requested plugins
        if args.plugin:
            try:
                document.Document.loadPlugins(pluginlist=args.plugin)
            except RuntimeError as e:
                startuperrors.append(str(e))

        # different modes
        if args.listen:
            # listen to incoming commands
            listen(args.docs, quiet=args.quiet)
        elif args.export:
            export(args.export, args.docs, args.export_option)
            self.quit()
            sys.exit(0)
        else:
            # standard start main window
            self.openMainWindow(args.docs)
            self.startupdone = True

        # clear splash when startup done
        if self.splash is not None:
            self.splash.finish(self.topLevelWidgets()[0])

        # this has to be displayed after the main window is created,
        # otherwise it never gets shown
        for error in startuperrors:
            qt.QMessageBox.critical(None, "Veusz", error)

    def showException(self, excepttype, exceptvalue, tracebackobj):
        """Show an exception dialog (raised from another thread)."""
        from veusz.dialogs.exceptiondialog import ExceptionDialog
        if not isinstance(exceptvalue, utils.IgnoreException):
            # next exception is ignored to clear out the stack frame of the
            # previous exception - yuck
            d = ExceptionDialog((excepttype, exceptvalue, tracebackobj), None)
            d.exec_()

def run():
    '''Run the main application.'''

    # high DPI support
    try:
        qt.QApplication.setAttribute(qt.Qt.AA_EnableHighDpiScaling, True)
        qt.QApplication.setAttribute(qt.Qt.AA_UseHighDpiPixmaps, True)
    except AttributeError:
        # old qt versions
        pass

    # jump to the embedding client entry point if required
    if len(sys.argv) == 2 and sys.argv[1] == '--embed-remote':
        from veusz.embed_remote import runremote
        runremote()
        return

    # start me up
    app = VeuszApp()
    app.startup()
    app.exec_()

# if ran as a program
if __name__ == '__main__':
    #import cProfile
    #cProfile.run('run()', 'outprofile.dat')
    run()



#    Copyright (C) 2013 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import numpy as N
from .. import qtall as qt
from .. import document
from .. import setting
from . import widget
from ..helpers import threed

def _(text, disambiguation=None, context='Graph3D'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

# non-reflective back of cube
class BackSurface(setting.Surface3D):
    def __init__(self, name, **args):
        setting.Surface3D.__init__(self, name, **args)
        self.get('color').newDefault('white')
        self.get('reflectivity').newDefault(0)
        self.get('hide').newDefault(True)

class Graph3D(widget.Widget):
    """3D graph (orthogonal axes) containing other plotting widgets."""

    typename='graph3d'
    allowusercreation = True
    description = _('3d graph')

    # start and end points of edges of cube
    _borderedges = (
        ((0,0,0), (0,0,1)),
        ((0,0,0), (0,1,0)),
        ((0,0,0), (1,0,0)),
        ((0,0,1), (0,1,1)),
        ((0,0,1), (1,0,1)),
        ((0,1,0), (0,1,1)),
        ((0,1,0), (1,1,0)),
        ((0,1,1), (1,1,1)),
        ((1,0,0), (1,0,1)),
        ((1,0,0), (1,1,0)),
        ((1,0,1), (1,1,1)),
        ((1,1,0), (1,1,1)),
    )

    # centres of each face
    _facecentres = (
        (0.0, 0.5, 0.5),
        (0.5, 0.0, 0.5),
        (0.5, 0.5, 0.0),
        (0.5, 0.5, 1.0),
        (0.5, 1.0, 0.5),
        (1.0, 0.5, 0.5),
    )

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        widget.Widget.addSettings(s)

        s.add( setting.Float(
            'xSize', 1.,
            minval=0.01, maxval=100,
            descr=_('X size'),
            usertext=_('X size') ))
        s.add( setting.Float(
            'ySize', 1.,
            minval=0.01, maxval=100,
            descr=_('Y size'),
            usertext=_('Y size') ))
        s.add( setting.Float(
            'zSize', 1.,
            minval=0.01, maxval=100,
            descr=_('Z size'),
            usertext=_('Z size') ))

        s.add( setting.Float(
            'xPos', 0.,
            minval=-100, maxval=100,
            descr=_('X position'),
            usertext=_('X position') ))
        s.add( setting.Float(
            'yPos', 0.,
            minval=-100, maxval=100,
            descr=_('Y position'),
            usertext=_('Y position') ))
        s.add( setting.Float(
            'zPos', 0.,
            minval=-100, maxval=100,
            descr=_('Z position'),
            usertext=_('Z position') ))

        s.add(setting.Line3D(
            'Border',
            descr = _('Graph border'),
            usertext = _('Border')),
            pixmap = 'settings_border' )
        s.add(BackSurface(
            'Back',
            descr = _('Graph back'),
            usertext = _('Back')),
            pixmap = 'settings_bgfill' )

    @classmethod
    def allowedParentTypes(self):
        from . import scene3d
        return (scene3d.Scene3D,)

    def addDefaultSubWidgets(self):
        """Add axes automatically."""
        from . import axis3d
        for n in ('x', 'y', 'z'):
            if self.parent.getChild(n) is None:
                ax = axis3d.Axis3D(self, name=n)
                ax.linkToStylesheet()

    def getAxesDict(self, axesnames, ignoremissing=False):
        """Get the axes for widgets to plot against.
        axesnames is a list/set of names to find.
        Returns a dict of objects
        """

        axes = {}
        # recursively go back up the tree to find axes
        w = self
        while w is not None and len(axes) < len(axesnames):
            for c in w.children:
                name = c.name
                if ( name in axesnames and name not in axes and
                     hasattr(c, 'isaxis3d') and c.isaxis3d ):
                    axes[name] = c
            w = w.parent

        # didn't find everything...
        if w is None and not ignoremissing:
            for name in axesnames:
                if name not in axes:
                    axes[name] = None

        # return list of found axes
        return axes

    def getAxes(self, axesnames):
        """Return a list of axes widgets given a list of names."""
        ad = self.getAxesDict(axesnames)
        return [ad[n] for n in axesnames]

    def addBorder(self, painter, root):
        s = self.settings
        if s.Border.hide:
            return
        lineprop = s.Border.makeLineProp(painter)
        edges = N.array(self._borderedges)
        ls = threed.LineSegments(
            threed.ValVector(N.ravel(edges[:,0,:])),
            threed.ValVector(N.ravel(edges[:,1,:])),
            lineprop)
        root.addObject(ls)

    def addBackSurface(self, painter, root):
        back = self.settings.Back
        if back.hide:
            return
        prop = back.makeSurfaceProp(painter)

        # triangles with the correct orientation of the norm vector
        # not to draw the surface if it is pointing towards the viewer
        for p1, p2, p3 in (
                ((0,0,0), (0,0,1), (1,0,0)),
                ((0,0,1), (0,0,0), (0,1,0)),
                ((0,1,0), (0,1,1), (0,0,1)),
                ((0,1,0), (1,1,0), (0,1,1)),
                ((0,1,0), (0,0,0), (1,0,0)),
                ((0,1,1), (1,0,1), (0,0,1)),
                ((0,1,1), (1,1,1), (1,0,1)),
                ((1,0,0), (1,1,0), (0,1,0)),
                ((1,0,1), (1,0,0), (0,0,1)),
                ((1,0,1), (1,1,0), (1,0,0)),
                ((1,0,1), (1,1,1), (1,1,0)),
                ((1,1,0), (1,1,1), (0,1,1)),
        ):
            root.addObject(threed.TriangleFacing(
                threed.Vec3(*p1), threed.Vec3(*p2), threed.Vec3(*p3), prop))

    def drawToObject(self, painter, painthelper):
        """Make objects, returning root"""

        s = self.settings

        # do no painting if hidden
        if s.hide:
            return

        # do axis min-max computation
        axestodraw = {}
        axesofwidget = painthelper.plotteraxismap

        for c in self.children:
            try:
                for a in axesofwidget[c]:
                    axestodraw[a.name] = a
            except (KeyError, AttributeError):
                if c.isaxis:
                    axestodraw[c.name] = c

        for axis in axestodraw.values():
            axis.computePlottedRange()

        cont = threed.ObjectContainer()
        cont.objM = (
            # graph position
            threed.translationM4(threed.Vec3(
                s.xPos - 0.5*s.xSize,
                s.yPos - 0.5*s.ySize,
                s.zPos - 0.5*s.zSize)) *
            # graph size
            threed.scaleM4(threed.Vec3(s.xSize, s.ySize, s.zSize))
        )

        # add graph box
        self.addBorder(painter, cont)
        # add graph behind fill
        self.addBackSurface(painter, cont)
        # make clickable
        cont.assignWidgetId(id(self))

        # reset counter and compute automatic colors
        painthelper.autoplottercount = 0
        for c in self.children:
            c.setupAutoColor(painter)

        # build 3d scene from children
        for c in self.children:
           obj = c.drawToObject(painter, painthelper)
           if obj:
               cont.addObject(obj)

        return cont

document.thefactory.register(Graph3D)



# utilfuncs.py
# utility functions

#    Copyright (C) 2003 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

import sys
import string
import re
import os
import os.path
import threading
import codecs
import io
import csv
import time
from collections import defaultdict

import numpy as N

from .. import qtall as qt

class IgnoreException(Exception):
    """A special exception class to be ignored by the exception handler."""

class InvalidType(Exception):
    """Exception used when invalid values are used in settings."""

def _getVeuszDirectory():
    """Get resource and examples directories for Veusz."""

    if hasattr(sys, 'frozen'):
        # for pyinstaller compatability
        resdir = os.path.dirname(os.path.abspath(sys.executable))
    else:
        # standard installation
        resdir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

    # override data directory with symlink
    if os.path.exists( os.path.join(resdir, 'resources') ):
        resdir = os.path.realpath( os.path.join(resdir, 'resources') )

    # override with VEUSZ_RESOURCE_DIR environment variable if necessary
    resdir = os.environ.get('VEUSZ_RESOURCE_DIR', resdir)

    # now get example directory (which may be a symlink)
    examplesdir = os.path.realpath( os.path.join(resdir, 'examples') )

    return resdir, examplesdir

# get resource and example directories
resourceDirectory, exampleDirectory = _getVeuszDirectory()

def getLicense():
    """Return license text."""
    try:
        f = open(os.path.join(resourceDirectory, 'COPYING'), 'rU')
        text = f.read()
        f.close()
    except EnvironmentError:
        text = (
            'Could not open the license file.\n'
            'See license at http://www.gnu.org/licenses/gpl-2.0.html'
        )
    return text

id_re = re.compile('^[A-Za-z_][A-Za-z0-9_]*$')
def validPythonIdentifier(name):
    """Is this a valid python identifier?"""
    return id_re.match(name) is not None

def validateDatasetName(name):
    """Validate dataset name is okay.
    Dataset names can contain anything except back ticks!
    """
    return len(name.strip()) > 0 and name.find('`') == -1

def validateWidgetName(name):
    """Validate widget name is okay.
    Widget names are valid if no surrounding whitespace and do not contain /
    """
    return ( len(name) > 0 and name.strip() == name and name.find('/') == -1
             and name != '.' and name != '..' )

def cleanDatasetName(name):
    """Make string into a valid dataset name."""
    # replace backticks and get rid of whitespace at ends
    return name.replace('`', '_').strip()

def relpath(filename, dirname):
    """Make filename a relative filename relative to dirname."""

    # spit up paths into components
    filename = os.path.abspath(filename)
    fileparts = filename.split(os.path.sep)
    dirparts = os.path.abspath(dirname).split(os.path.sep)

    # if first non empty part is non equal, return original
    i = 0
    while not fileparts[i] and not dirparts[i]:
        i += 1
    if fileparts[i] != dirparts[i]:
        return filename

    # remove equal bits at start
    while fileparts and dirparts and fileparts[0] == dirparts[0]:
        fileparts.pop(0)
        dirparts.pop(0)

    # add on right number of .. to get back up
    fileparts = [os.path.pardir]*len([d for d in dirparts if d]) + fileparts

    # join parts back together
    return os.path.sep.join(fileparts)

def extendedColorFromQColor(col):
    """Make an extended color #RRGGBBAA or #RRGGBB string."""
    if col.alpha() == 255:
        return str(col.name())
    else:
        return '#%02x%02x%02x%02x' % (
            col.red(), col.green(), col.blue(), col.alpha())

def pixmapAsHtml(pix):
    """Get QPixmap as html image text."""
    ba = qt.QByteArray()
    buf = qt.QBuffer(ba)
    buf.open(qt.QIODevice.WriteOnly)
    pix.toImage().save(buf, "PNG")
    b64 = bytes(buf.data().toBase64()).decode('ascii')
    return '<img src="data:image/png;base64,%s">' % b64

def pythonise(text):
    """Turn an expression of the form 'A b c d' into 'A(b,c,d)'.

    This is for 'pythonising' commands from a command-line interface
    to make it easier for users. We also have to take account of quotes
    and backslashes.
    """

    out = ''
    insingle = False    # in a single quote section
    indouble = False    # in a double quote section
    firstnonws = False  # have we reached first non WS char
    firstpart = True    # have we appended the first part of the expr
    lastbslash = False  # was the last character a back-slash

    # iterate over characters
    for c in text:

        # keep leading WS
        if c in string.whitespace and not firstnonws:
            out += c
            continue
        firstnonws = True

        # this character isn't escaped
        if not lastbslash:

            # quoted section
            if c == "'":
                insingle = not insingle
            elif c == '"':
                indouble = not indouble

            elif c == '\\':
                lastbslash = True
                continue

            # spacing between parts
            if c == ' ' and not insingle and not indouble:
                if firstpart:
                    out += '('
                    firstpart = False
                else:
                    out += ','
            else:
                out += c
        else:
            out += '\\' + c
            lastbslash = False

    # we're still in the first part
    if firstpart:
        out += '('

    # we can add a right bracket
    if not insingle and not indouble:
        out += ')'

    return out

def validLinePoints(x, y):
    """Take x and y points and split into sets of points which
    don't have invalid points.
    This is a generator.
    """
    xvalid = N.logical_not( N.isfinite(x) ).nonzero()[0]
    yvalid = N.logical_not( N.isfinite(y) ).nonzero()[0]
    invalid = N.concatenate((xvalid, yvalid))
    invalid.sort()
    last = 0
    for valid in invalid:
        if valid > last:
            yield x[last:valid], y[last:valid]
        last = valid + 1
    if last < x.shape[0]-1:
        yield x[last:], y[last:]

class NonBlockingReaderThread(threading.Thread):
    """A class to read blocking file objects and return the result.

    Usage:
     r = ReadThread(myfile)
     r.start()
     while True:
      newdata, done = r.getNewData()
      print newdata
      if done: break

    This is used mainly because windows doesn't properly support
    non-blocking pipes as files.

    If exiteof is True, then exit capturing when we can capture no
    more data.
    """

    def __init__(self, fileobject, exiteof=True):
        """Create the thread object."""
        threading.Thread.__init__(self)
        self.fileobject = fileobject
        self.lock = threading.Lock()
        self.data = ''
        self.done = False
        self.exiteof = exiteof

    def getNewData(self):
        """Get any data waiting to be read, and whether
        the reading is finished.

        Returns (data, done)
        """
        self.lock.acquire()
        data = self.data
        done = self.done
        self.data = ''
        self.lock.release()
        if isinstance(data, Exception):
            # if the reader errored somewhere
            raise data
        else:
            return data, done

    def run(self):
        """Do the reading from the file object."""

        while True:
            try:
                data = self.fileobject.readline()
            except Exception as e:
                # error in reading
                self.lock.acquire()
                self.data = e
                self.lock.release()
                break

            # no more data: end of file
            if len(data) == 0:
                if self.exiteof:
                    self.lock.acquire()
                    self.done = True
                    self.lock.release()
                    break
                else:
                    time.sleep(0.1)
            else:
                self.lock.acquire()
                self.data += data
                self.lock.release()

# standard python encodings
encodings = [
    'ascii', 'big5hkscs', 'big5', 'charmap', 'cp037', 'cp424',
    'cp437', 'cp500', 'cp737', 'cp775', 'cp850', 'cp852', 'cp855',
    'cp856', 'cp857', 'cp860', 'cp861', 'cp862', 'cp863', 'cp864',
    'cp865', 'cp866', 'cp869', 'cp874', 'cp875', 'cp932', 'cp949',
    'cp950', 'cp1006', 'cp1026', 'cp1140','cp1250', 'cp1251', 'cp1252',
    'cp1253', 'cp1254', 'cp1255', 'cp1256', 'cp1257', 'cp1258',
    'euc_jis_2004', 'euc_jisx0213', 'euc_jp', 'euc_kr', 'gb18030', 'gb2312',
    'gbk', 'hp_roman8', 'hz', 'iso2022_jp_1', 'iso2022_jp_2004',
    'iso2022_jp_2', 'iso2022_jp_3', 'iso2022_jp_ext', 'iso2022_jp',
    'iso2022_kr', 'iso8859_10', 'iso8859_11', 'iso8859_13', 'iso8859_14',
    'iso8859_15', 'iso8859_16', 'iso8859_1', 'iso8859_2', 'iso8859_3',
    'iso8859_4', 'iso8859_5', 'iso8859_6', 'iso8859_7', 'iso8859_8',
    'iso8859_9', 'johab', 'koi8_r', 'koi8_u', 'latin_1', 'mac_arabic',
    'mac_centeuro', 'mac_croatian', 'mac_cyrillic', 'mac_farsi', 'mac_greek',
    'mac_iceland', 'mac_latin2', 'mac_romanian', 'mac_roman', 'mac_turkish',
    'ptcp154', 'shift_jis_2004', 'shift_jis', 'shift_jisx0213',
    'tis_620','utf_16_be', 'utf_16_le', 'utf_16',
    'utf_32_be', 'utf_32_le', 'utf_32', 'utf_7', 'utf_8', 'utf_8_sig'
]

def openEncoding(filename, encoding, mode='r'):
    """Convenience function for opening file with encoding given.

    If filename == '{clipboard}', then load the data from the clipboard
    instead.
    """
    if filename == '{clipboard}':
        text = qt.QApplication.clipboard().text()
        return io.StringIO(text)
    else:
        return io.open(filename, mode, encoding=encoding, errors='ignore')

# The following two classes are adapted from the Python documentation
# they are modified to turn off encoding errors

class _UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8

    Needed for python2
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f, errors='ignore')
    def __iter__(self):
        return self
    def next(self):
        line = self.reader.next()
        return line.encode("utf-8")

class _UTF8Decoder:
    """
    Python2 iterator than decodes lists of utf-8 encoded strings
    """
    def __init__(self, iterator):
        self.iterator = iterator
    def __iter__(self):
        return self
    def next(self):
        line = self.iterator.next()
        return [str(x, "utf-8") for x in line]

def get_unicode_csv_reader(filename, dialect=csv.excel,
                           encoding='utf-8', **kwds):
    """Return an iterator to iterate over CSV file with encoding given."""

    if filename != '{clipboard}':
        f = open(filename, encoding=encoding, errors='ignore')
    else:
        # take the unicode clipboard and just put into utf-8 format
        s = qt.QApplication.clipboard().text()
        f = io.StringIO(s)

    reader = csv.reader(f, dialect=dialect, **kwds)
    return reader

# End python doc classes

def populateCombo(combo, items):
    """Populate the combo with the list of items given.

    This also makes sure the currently entered text persists,
    or if currenttext is set, use this
    """

    # existing setting
    currenttext = combo.currentText()

    # add to list if not included
    if currenttext not in items:
        items = items + [currenttext]

    # put in new entries
    for i, val in enumerate(items):
        if i >= combo.count():
            combo.addItem(val)
        else:
            if combo.itemText(i) != val:
                combo.insertItem(i, val)

    # remove any extra items
    while combo.count() > len(items):
        combo.removeItem( combo.count()-1 )

    # get index for current value
    index = combo.findText(currenttext)
    combo.setCurrentIndex(index)

def positionFloatingPopup(popup, widget):
    """Position a popped up window (popup) to side and below widget given."""
    pos = widget.parentWidget().mapToGlobal( widget.pos() )
    desktop = qt.QApplication.desktop()

    # recalculates out position so that size is correct below
    popup.adjustSize()

    # is there room to put this widget besides the widget?
    if pos.y() + popup.height() + 1 < desktop.height():
        # put below
        y = pos.y() + 1
    else:
        # put above
        y = pos.y() - popup.height() - 1

    # is there room to the left for us?
    if ( (pos.x() + widget.width() + popup.width() < desktop.width()) or
         (pos.x() + widget.width() < desktop.width()/2) ):
        # put left justified with widget
        x = pos.x() + widget.width()
    else:
        # put extending to left
        x = pos.x() - popup.width() - 1

    popup.move(x, y)
    popup.setFocus()

# based on http://stackoverflow.com/questions/10607841/algorithm-for-topological-sorting-if-cycles-exist
def topological_sort(dependency_pairs):
    """Given a list of pairs, perform a topological sort.
    That means, each item has something which needs to be done first.
    topsort( [(1,2), (3,4), (5,6), (1,3), (1,5), (1,6), (2,5)] )
    returns  [1, 2, 3, 5, 4, 6], []
    for ordered and cyclic items
    """

    num_heads = defaultdict(int)   # num arrows pointing in
    tails = defaultdict(list)      # list of arrows going out
    for h, t in dependency_pairs:
        num_heads[t] += 1
        tails[h].append(t)

    ordered = [h for h in tails if h not in num_heads]
    i = 0
    while i < len(ordered):
        h = ordered[i]
        for t in tails[h]:
            num_heads[t] -= 1
            if not num_heads[t]:
                ordered.append(t)
        i += 1
    cyclic = [n for n, heads in num_heads.items() if heads]
    return ordered, cyclic

def isiternostr(i):
    """Is this iterator, but not a string?"""
    return hasattr(i, '__iter__') and not isinstance(i, str)

def nextfloat(fin):
    """Return (approximately) next float value (for f>0)."""
    d = 2**-52
    split = N.frexp(fin)
    while True:
        fout = N.ldexp(split[0] + d, split[1])
        if fin != fout:
            return fout
        d *= 2

def round2delt(fin1, fin2):
    """Take two float values. Return value rounded to number
    of decimal places where they differ."""

    if not N.isfinite(fin1) or not N.isfinite(fin2):
        return fin1

    # round up to next value to avoid 0.999999...
    f1 = nextfloat(abs(fin1))
    f2 = nextfloat(abs(fin2))

    maxlog = int( max(N.log10(f1), N.log10(f2)) + 1 )
    # note: out2 unused, but useful for debugging
    if maxlog < 0:
        out1 = out2 = '0.' + '0'*(-1-maxlog)
    else:
        out1 = out2 = ''

    for i in range(maxlog,-200,-1):
        p = 10**i
        d1, d2 = int(f1/p), int(f2/p)
        f1 -= int(d1)*p
        f2 -= int(d2)*p

        c1 = chr(d1 + 48) # 48 == '0'
        c2 = chr(d2 + 48)
        out1 += c1
        out2 += c2

        if c1 != c2 and p < abs(fin1): # at least 1 sig fig
            if i > 0:
                # add missing zeros
                out1 += '0'*i
                out2 += '0'*i
            break

        if i == 0:
            out1 += '.'
            out2 += '.'

    # convert back to float for output
    fout = float(out1)
    return fout if fin1 > 0 else -fout

def checkOrder(inv):
    """Check order of inv
    Returns: +1: ascending order
             -1: descending order
              0: other, or non finite."""
    v = N.array(inv)
    if not N.all(N.isfinite(v)):
        return 0
    delta = v[1:] - v[:-1]
    if N.all(delta > 0):
        return 1
    if N.all(delta < 0):
        return -1
    return 0

def checkAscending(v):
    """Check list of values is finite and ascending."""
    v = N.array(v)
    if not N.all( N.isfinite(v) ):
        return False
    return N.all( (v[1:] - v[:-1]) > 0 )

def rrepr(val):
    """Reproducible repr.

    The idea is to make a repr which won't change. We sort dict and
    set entries."""

    if isinstance(val, dict):
        l = [ "%s: %s" % (rrepr(k), rrepr(val[k]))
              for k in sorted(val) ]
        return "{%s}" % ", ".join(l)
    elif isinstance(val, set):
        l = [rrepr(v) for v in sorted(val)]
        return "set([%s])" % ", ".join(l)
    elif isinstance(val, list):
        l = [rrepr(v) for v in val]
        return "[%s]" % ", ".join(l)
    else:
        return repr(val)

def escapeHDFDataName(name):
    """Return escaped dataset name for saving in HDF5 files.
    This is because names cannot include / characters in HDF5
    """
    name = name.replace('`', '`BT')
    name = name.replace('/', '`SL')
    return name.encode('utf-8')

def unescapeHDFDataName(name):
    """Return original name after being escaped."""
    name = name.replace('`SL', '/')
    name = name.replace('`BT', '`')
    return name

def allNotNone(*items):
    """Are all the items not None."""
    return not any((x is None for x in items))

def anyNone(*items):
    """Are any items None."""
    return any((x is None for x in items))

def findOnPath(cmd):
    """Find a command on the system path, or None if does not exist."""
    path = os.getenv('PATH', os.path.defpath)
    pathparts = path.split(os.path.pathsep)
    for dirname in pathparts:
        dirname = dirname.strip('"')
        cmdtry = os.path.join(dirname, cmd)
        if os.path.isfile(cmdtry) and os.access(cmdtry, os.X_OK):
            return cmdtry
    return None

def listIndex(inlist, item):
    """Return index of item in list or -1 if not available."""
    try:
        return inlist.index(item)
    except ValueError:
        return -1

class Struct:
    """Simple structure-like class."""

    def __init__(self, **args):
        self.__dict__.update(args)

    def __repr__(self):
        return '<%s>' % str(
            ', '.join(
                '%s:%s' % (k, repr(getattr(self, k)))
                for k in sorted(self.__dict__)
            )
        )

class SvgWidgetFixedAspect(qt.QWidget):
    """Draw an SVG file with the aspect ratio fixed to the original."""

    def __init__(self, filename, *args):
        qt.QWidget.__init__(self, *args)
        self.renderer = qt.QSvgRenderer(filename, self)
        self.defwidth = self.renderer.defaultSize().width()
        self.defheight = self.renderer.defaultSize().height()

    def sizeHint(self):
        return qt.QSize(self.defwidth, self.defheight)

    def paintEvent(self, evt):
        painter = qt.QPainter(self)

        w = self.width()
        h = self.height()

        ratio_x = w / self.defwidth
        ratio_y = h / self.defheight
        if ratio_x < ratio_y:
            outw = w
            outh = w*self.defheight//self.defwidth
            left = 0
            top = (h-outh)//2
        else:
            outw = h*self.defwidth//self.defheight
            outh = h
            left = (w-outw)//2
            top = 0

        self.renderer.render(painter, qt.QRectF(left, top, outw, outh))

class OverrideCursor:
    """A context manager to handle changing the mouse cursor temporarily."""

    def __init__(self, cursor=qt.Qt.WaitCursor):
        self.cursor = cursor

    def __enter__(self):
        qt.QApplication.setOverrideCursor(qt.QCursor(self.cursor))

    def __exit__(self, typ, value, traceback):
        qt.QApplication.restoreOverrideCursor()

class DisabledIconEngine(qt.QIconEngine):
    """Icon engine which draws icons in a disabled state."""

    def __init__(self, icon):
        qt.QIconEngine.__init__(self)
        self.icon = icon

    def actualSize(self, size, mode, state):
        return size

    def clone(self):
        return GreyIconEngine(self.icon)

    def iconName(self):
        return ''

    def key(self):
        return str(id(self))

    def paint(self, painter, rect, mode, state):
        self.icon.paint(painter, rect, qt.Qt.AlignCenter, qt.QIcon.Disabled, state)

    def pixmap(self, size, mode, state):
        return self.icon.pixmap(size, qt.QIcon.Disabled, state)
